# ---------------------------------------------------------------------------
# Writer to shared storage devices
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import bz2
import queue
import threading
import mtda.constants as CONSTS
import zlib
import zstandard as zstd
import lzma


class AsyncImageWriter(queue.Queue):

    def __init__(self, mtda, storage, compression=CONSTS.IMAGE.RAW):
        self.mtda = mtda
        self.storage = storage
        self.compression = compression
        self._blksz = CONSTS.WRITER.WRITE_SIZE
        self._exiting = False
        self._failed = False
        self._thread = None
        self._writing = False
        self._written = 0
        self._zdec = None
        super().__init__(maxsize=CONSTS.WRITER.QUEUE_SLOTS)

    @property
    def compression(self):
        self.mtda.debug(3, "storage.writer.compression.get()")

        result = self._compression

        self.mtda.debug(3, "storage.writer.compression.get(): "
                           "%s" % str(result))
        return result

    @compression.setter
    def compression(self, compression):
        self.mtda.debug(3, "storage.writer.compression.set()")

        compression = CONSTS.IMAGE(compression)
        if compression == CONSTS.IMAGE.RAW:
            self._write = self.write_raw
        elif compression == CONSTS.IMAGE.BZ2:
            self._write = self.write_bz2
        elif compression == CONSTS.IMAGE.GZ:
            self._write = self.write_gz
        elif compression == CONSTS.IMAGE.ZST:
            self._write = self.write_zst
        elif compression == CONSTS.IMAGE.XZ:
            self._write = self.write_xz
        else:
            raise ValueError("unsupported image compression!")
        self._compression = compression

        result = compression
        self.mtda.debug(3, "storage.writer.compression.set(): "
                           "%s" % str(result))

    @property
    def failed(self):
        return self._failed

    def put(self, chunk, block=True, timeout=None):
        self.mtda.debug(3, "storage.writer.put()")

        if self.storage is None:
            self.mtda.debug(1, "storage.writer.put(): no storage!")
            raise IOError("no storage!")
        result = super().put(chunk, block, timeout)

        self.mtda.debug(3, "storage.writer.put(): %s" % str(result))
        return result

    def start(self):
        self.mtda.debug(3, "mtda.storage.writer.start()")

        result = None
        self._thread = threading.Thread(target=self.worker,
                                        daemon=True, name='writer')
        self._thread.start()

        self.mtda.debug(3, "storage.writer.start(): %s" % str(result))
        return result

    def stop(self):
        self.mtda.debug(3, "storage.writer.stop()")

        result = None
        self.mtda.debug(2, "storage.writer.stop(): waiting on queue...")
        self.join()

        if self._thread is not None:
            self.mtda.debug(2, "storage.writer.stop(): waiting on thread...")
            self._exiting = True
            self.put(b'')
            self._thread.join()

        self.mtda.debug(2, "storage.writer.stop(): all done")
        self._thread = None
        self._zdec = None

        self.mtda.debug(3, "storage.writer.stop(): %s" % str(result))
        return result

    def worker(self):
        self.mtda.debug(3, "storage.writer.worker()")

        result = None
        self._exiting = False
        self._failed = False
        self._written = 0
        while self._exiting is False:
            chunk = self.get()
            if self._exiting is False:
                self._writing = True
                self._write(chunk)
                self._writing = False
            self.task_done()
            if self._failed is True:
                self.mtda.debug(1, "storage.writer.worker(): "
                                   "write or decompression error!")
                break

        self.mtda.debug(3, "storage.writer.worker(): %s" % str(result))
        return result

    def write_raw(self, data, session=None):
        self.mtda.debug(3, "storage.writer.write_raw()")

        result = None
        try:
            result = self.storage.write(data)
            self._written += result if result is not None else 0
        except OSError as e:
            self.mtda.debug(1, "storage.writer.write_raw(): "
                               "%s" % str(e.args[0]))
            self._failed = True

        self.mtda.debug(3, "storage.writer.write_raw(): %s" % str(result))
        return result

    def write_gz(self, data, session=None):
        self.mtda.debug(3, "storage.writer.write_gz()")

        # Create a zlib decompressor when called for the first time
        if self._zdec is None:
            self._zdec = zlib.decompressobj(16+zlib.MAX_WBITS)

        try:
            cont = True
            result = None
            while cont is True:
                uncompressed = self._zdec.decompress(data, self._blksz)
                data = self._zdec.unconsumed_tail
                cont = len(data) > 0
                result = self.storage.write(uncompressed)
                self._written += result if result is not None else 0
        except (OSError, zlib.error) as e:
            self.mtda.debug(1, "storage.writer.write_gz(): "
                               "%s" % str(e.args[0]))
            self._failed = True

        self.mtda.debug(3, "storage.writer.write_gz(): %s" % str(result))
        return result

    def write_bz2(self, data):
        self.mtda.debug(3, "storage.writer.write_bz2()")

        # Create a bz2 decompressor when called for the first time
        if self._zdec is None:
            self._zdec = bz2.BZ2Decompressor()

        try:
            cont = True
            result = None
            while cont is True:
                uncompressed = self._zdec.decompress(data, self._blksz)
                result = self.storage.write(uncompressed)
                self._written += result if result is not None else 0
                cont = self._zdec.needs_input is False
                data = b''
        except EOFError:
            result = 0
        except OSError as e:
            self.mtda.debug(1, "storage.writer.write_bz2(): "
                               "%s" % str(e.args[0]))
            self._failed = True

        self.mtda.debug(3, "storage.writer.write_bz2(): %s" % str(result))
        return result

    def write_zst(self, data):
        self.mtda.debug(3, "storage.writer.write_zst()")

        # Create a decompressor when called for the first time
        if self._zdec is None:
            dctx = zstd.ZstdDecompressor()
            self._zdec = dctx.stream_writer(self.storage)
        try:
            result = self._zdec.write(data)
            self._written += result if result is not None else 0
        except OSError:
            self.mtda.debug(1, "storage.writer.write_zst(): write error!")
            self._failed = True

        self.mtda.debug(3, "storage.writer.write_zst(): %s" % str(result))
        return result

    def write_xz(self, data):
        self.mtda.debug(3, "storage.writer.write_xz()")

        # Create a xz decompressor when called for the first time
        if self._zdec is None:
            self._zdec = lzma.LZMADecompressor()

        try:
            cont = True
            result = None
            while cont is True:
                uncompressed = self._zdec.decompress(data, self._blksz)
                result = self.storage.write(uncompressed)
                self._written += result if result is not None else 0
                cont = self._zdec.needs_input is False
                data = b''
        except EOFError:
            result = 0
        except OSError as e:
            self.mtda.debug(1, "storage.writer.write_xz(): "
                               "%s" % str(e.args[0]))
            self._failed = True

        self.mtda.debug(3, "storage.writer.write_xz(): %s" % str(result))
        return result

    @property
    def writing(self):
        return self._writing

    @property
    def written(self):
        return self._written



# ---------------------------------------------------------------------------
# Console logger for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import codecs
from collections import deque
import sys
import threading
import time


class ConsoleLogger:

    def __init__(self, mtda, console, socket=None,
                 power_controller=None, topic=b'CON',
                 www=None):
        self.mtda = mtda
        self.console = console
        self._prompt = "=> "
        self.power_controller = power_controller
        self.prints = True
        self.rx_active = threading.Event()
        self.rx_alive = False
        self.rx_thread = None
        self.rx_queue = bytearray()
        self.rx_buffer = deque(maxlen=1000)
        self.rx_lock = threading.Lock()
        self.rx_cond = threading.Condition(self.rx_lock)
        self.socket = socket
        self.topic = topic
        self.basetime = 0
        self.timestamps = False
        self._time_from = None
        self._time_until = None
        self._www = www

    def start(self):
        self.rx_alive = True
        self.rx_thread = threading.Thread(
            target=self.reader, name='console_rx')
        self.rx_thread.daemon = True
        self.rx_thread.start()

    def stop(self):
        self.pause()
        with self.rx_lock:
            self.rx_alive = False
            self.rx_cond.notify()

    def _clear(self):
        self.rx_buffer.clear()
        self.rx_queue = bytearray()

    def clear(self):
        with self.rx_lock:
            self._clear()

    def _dump(self, flush=True):
        data = ""
        lines = len(self.rx_buffer)
        while lines > 0:
            line = self.rx_buffer.popleft().decode("utf-8", "ignore")
            data = data + line
            lines = lines - 1
        line = self.rx_queue.decode("utf-8", "ignore")
        data = data + line
        if flush:
            self.rx_queue = bytearray()
        return data

    def dump(self):
        with self.rx_lock:
            data = self._dump(flush=False)
        return data

    def flush(self):
        with self.rx_lock:
            data = self._dump(flush=True)
        return data

    def _head(self):
        if len(self.rx_buffer) > 0:
            line = self.rx_buffer.popleft().decode("utf-8", "ignore")
        else:
            line = None
        return line

    def head(self):
        with self.rx_lock:
            line = self._head()
        return line

    def lines(self):
        with self.rx_lock:
            lines = len(self.rx_buffer)
        return lines

    def _matchprompt(self):
        prompt = self._tail(False)
        if prompt is None:
            return False
        if prompt.startswith("\r"):
            prompt = prompt[1:]
        return prompt.endswith(self._prompt)

    def prompt(self, newPrompt=None):
        with self.rx_lock:
            if newPrompt is not None:
                self._prompt = newPrompt
            p = self._prompt
        return p

    def run(self, cmd):
        self.rx_lock.acquire()
        self._clear()

        # Send a break to get a prompt
        self.write("\3")

        # Wait for a prompt
        self.rx_cond.wait_for(self._matchprompt)

        # Send requested command
        self._clear()
        self.write("%s\n" % (cmd))

        # Wait for the command to complete
        self.rx_cond.wait_for(self._matchprompt)

        # Strip first line (command we sent) and flush received bytes
        self._head()
        data = self._dump(flush=True)

        # Release and return command output
        self.rx_lock.release()
        return data

    def _tail(self, discard=True):
        if len(self.rx_queue) > 0:
            line = self.rx_queue
        elif len(self.rx_buffer) > 0:
            line = self.rx_buffer[-1]
        else:
            return None
        if discard is True:
            self._clear()
        return line.decode("utf-8", "ignore")

    def tail(self):
        with self.rx_lock:
            line = self._tail()
        return line

    def _match_any(self):
        self.mtda.debug(3, "console.logger._match_any()")

        result = False
        line = self._tail(False)
        if line is not None and self._what in line:
            self.mtda.debug(2, "matched '%s'" % str(self._what))
            result = True

        self.mtda.debug(3, "console.logger._match_any: %s" % str(result))
        return result

    def wait(self, what, timeout=None):
        self.mtda.debug(3, "console.logger.wait()")
        self.rx_lock.acquire()

        # Wait for specified string
        self._what = what
        result = self.rx_cond.wait_for(self._match_any, timeout)

        # Matched?
        if result is True:
            self.rx_lock.release()

        self.mtda.debug(3, "console.logger.wait: %s" % str(result))
        return result

    def write(self, data, raw=False):
        try:
            if raw is False:
                data = codecs.escape_decode(bytes(data, "utf-8"))[0]
            else:
                data = bytes(data, "utf-8")
            self.console.write(data)
        except IOError as e:
            print("write error on the console ({0})!".format(
                e.strerror), file=sys.stderr)

    def reset_timer(self):
        self.basetime = 0

    def toggle_prints(self):
        self.prints = not self.prints

    def toggle_timestamps(self):
        self.timestamps = not self.timestamps

    # Print bytes to the console (local or remote)
    def _print(self, data):
        if self.prints is True:
            if self.socket is not None:
                self.mtda.publish(self.topic, data)
            else:
                # Write to stdout if received are not pushed to the network
                sys.stdout.buffer.write(data)
                sys.stdout.buffer.flush()
            if self._www is not None:
                self._www.write(self.topic, data.decode('utf-8', 'ignore'))

    # Print a string to the console (local or remote)
    def print(self, data):
        if self.prints is True:
            data = codecs.escape_decode(bytes(data, "utf-8"))[0]
            self._print(data)

    def process_rx(self, data):
        # Initialize basetime on the 1st byte we receive
        if not self.basetime:
            self.basetime = time.time()

        # Add timestamps
        if self.timestamps is True or self._time_from is not None:
            newdata = bytearray()
            linefeeds = 0
            now = time.time()
            if self._time_from is not None and data.count(self._time_from):
                self.basetime = now
                self.timestamps = True
            for x in data:
                if x == 0xd:
                    continue
                newdata.append(x)
                if x == 0xa:
                    newdata.extend(b'\r')
                    if self.timestamps is True:
                        elapsed = now - self.basetime
                        timestr = "[%4.6f] " % elapsed
                        newdata.extend(timestr.encode("utf-8"))
                    linefeeds = linefeeds + 1
            data = newdata
            if self._time_until is not None and data.count(self._time_until):
                self.timestamps = False
        else:
            linefeeds = 1

        # Publish received data
        self._print(data)

        # Prevent concurrent access to the RX buffers
        self.rx_lock.acquire()

        # Add received data
        self.rx_queue.extend(data)

        # Find lines we have in the queue
        while linefeeds > 0:
            sz = len(self.rx_queue)
            off = self.rx_queue.find(b'\n', 0)
            if off >= 0:
                # Will include the line feed character in the buffered line
                off = off + 1
                rem = sz - off

                # Extract line from the RX queue
                line = self.rx_queue[:off]

                # Strip trailing \r
                if len(line) > 1 and line[-2] == 0xd and line[-1] == 0xa:
                    del line[-1:]
                    line[-1] = 0xa

                # Add this line to the circular buffer
                self.rx_buffer.append(line)

                # Remove consumed bytes from the queue
                if rem > 0:
                    self.rx_queue = self.rx_queue[-rem:]
                else:
                    self.rx_queue = bytearray()
            else:
                linefeeds = 0

        # Notify threads waiting on data
        self.rx_cond.notify()

        # Release access to the RX buffers
        self.rx_lock.release()

    def reader(self):
        con = self.console
        error = None
        retries = 3

        while self.rx_alive is True:
            self.rx_active.wait()

            try:
                data = con.read(con.pending() or 1)
                self.process_rx(data)
                continue
            except (IOError, TypeError):
                error = sys.exc_info()[0]

            try:
                if retries > 0:
                    print("resetting console to recover from read error "
                          "({0})...".format(error), file=sys.stderr)
                    con.close()
                    self.rx_active.wait()
                    con.open()
                    error = None
                else:
                    print("failed to reset the console, "
                          "aborting!", file=sys.stderr)
                    self.rx_alive = False
            except IOError:
                retries = retries - 1

    def pause(self):
        self.mtda.debug(3, "console.logger.pause()")

        with self.rx_lock:
            self.rx_active.clear()
            result = self.console.close()

        self.mtda.debug(3, "console.logger.pause(): "
                           "{}".format(result))
        return result

    def resume(self):
        self.mtda.debug(3, "console.logger.resume()")

        with self.rx_lock:
            result = self.console.open()
            if result is True:
                self.rx_active.set()

        self.mtda.debug(3, "console.logger.resume(): "
                           "{}".format(result))
        return result

    @property
    def time_from(self):
        if self._time_from is None:
            result = ""
        else:
            result = self._time_from.encode("utf-8")
        return result

    @time_from.setter
    def time_from(self, from_str):
        if from_str is None:
            self._time_from = None
        else:
            self._time_from = bytes(from_str, 'utf-8')

    @property
    def time_until(self):
        if self._time_until is None:
            result = ""
        else:
            result = self._time_until.encode("utf-8")
        return result

    @time_until.setter
    def time_until(self, until_str):
        if until_str is None:
            self._time_until = None
        else:
            self._time_until = bytes(until_str, 'utf-8')



# ---------------------------------------------------------------------------
# QEMU console driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import array
import fcntl
import termios

# Local imports
from mtda.console.interface import ConsoleInterface


class QemuConsole(ConsoleInterface):

    def __init__(self, mtda):
        self.mtda = mtda
        self.qemu = mtda.power_controller
        self.opened = False

    """ Configure this console from the provided configuration"""
    def configure(self, conf, role='console'):
        self.mtda.debug(3, "console.qemu.configure()")

    def probe(self):
        self.mtda.debug(3, "console.qemu.probe()")
        result = self.qemu.variant == "qemu"
        self.mtda.debug(3, "console.qemu.probe(): %s" % str(result))
        return result

    def open(self):
        self.mtda.debug(3, "console.qemu.open()")

        result = self.opened
        if self.opened is False:
            try:
                self.tx = open("/tmp/qemu-serial.in",  mode="wb", buffering=0)
                self.rx = open("/tmp/qemu-serial.out", mode="rb", buffering=0)

                result = True
            finally:
                self.opened = result
        else:
            self.mtda.debug(4, "console.qemu.open(): already opened")

        self.mtda.debug(3, "console.qemu.open(): %s" % str(result))
        return result

    def close(self):
        self.mtda.debug(3, "console.qemu.close()")

        result = True

        self.mtda.debug(3, "console.qemu.close(): %s" % str(result))
        return result

    """ Return number of pending bytes to read"""
    def pending(self):
        self.mtda.debug(3, "console.qemu.pending()")

        result = 0
        if self.opened is True:
            avail = array.array('l', [0])
            fcntl.ioctl(self.rx, termios.FIONREAD, avail, 1)
            result = avail[0]

        self.mtda.debug(3, "console.qemu.pending(): %s" % str(result))
        return result

    """ Read bytes from the console"""
    def read(self, n=1):
        self.mtda.debug(3, "console.qemu.read()")

        result = None
        if self.opened is True:
            try:
                result = self.rx.read(n)
            except BlockingIOError:
                result = None

        if result is None:
            result = b''

        self.mtda.debug(3, "console.qemu.read(): %s" % str(result))
        return result

    """ Write to the console"""
    def write(self, data):
        self.mtda.debug(3, "console.qemu.write(data=%s)" % str(data))

        result = None
        if self.opened is True:
            result = self.tx.write(data)

        self.mtda.debug(3, "console.qemu.write(): %s" % str(result))
        return result


def instantiate(mtda):
    return QemuConsole(mtda)



# ---------------------------------------------------------------------------
# Web service for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

from flask import Flask, render_template, request, session, send_from_directory
from flask_socketio import SocketIO
from urllib.parse import urlparse

import secrets
import threading
import uuid

import mtda.constants as CONSTS

app = Flask("mtda")
app.config["mtda"] = None
socket = SocketIO(app)


@app.route("/")
def index():
    return render_template("index.html")


@app.route('/assets/<path:path>')
def assets_dir(path):
    return send_from_directory('assets', path)


@app.route('/novnc/<path:path>')
def novnc_dir(path):
    return send_from_directory('/usr/share/novnc', path)


@socket.on("connect", namespace="/mtda")
def connect():
    session['id'] = uuid.uuid4().hex
    mtda = app.config['mtda']
    if mtda is not None:
        version = mtda.agent_version()
        socket.emit("mtda-version", {"version": version}, namespace="/mtda")

        data = mtda.console_dump()
        socket.emit("console-output", {"output": data}, namespace="/mtda")

        if mtda.video is not None:
            fmt = mtda.video.format
            url = urlparse(request.base_url)
            url = mtda.video.url(host=url.hostname)
            info = {"format": fmt, "url": url}
            socket.emit("video-info", info, namespace="/mtda")


@socket.on("console-input", namespace="/mtda")
def console_input(data):
    sid = session_id()
    mtda = app.config['mtda']
    if mtda is not None:
        mtda.console_send(data['input'], raw=False, session=sid)


@app.route('/keyboard-input')
def keyboard_input():
    mtda = app.config['mtda']
    map = {
      "Esc": mtda.keyboard.esc,
      "F1": mtda.keyboard.f1,
      "F2": mtda.keyboard.f2,
      "F3": mtda.keyboard.f3,
      "F4": mtda.keyboard.f4,
      "F5": mtda.keyboard.f5,
      "F6": mtda.keyboard.f6,
      "F7": mtda.keyboard.f7,
      "F8": mtda.keyboard.f8,
      "F9": mtda.keyboard.f9,
      "F10": mtda.keyboard.f10,
      "F11": mtda.keyboard.f11,
      "F12": mtda.keyboard.f12,
      "Backspace": mtda.keyboard.backspace,
      "Tab": mtda.keyboard.tab,
      "Caps Lock": mtda.keyboard.capsLock,
      "Enter": mtda.keyboard.enter,
      "Left": mtda.keyboard.left,
      "Right": mtda.keyboard.right,
      "Up": mtda.keyboard.up,
      "Down": mtda.keyboard.down
    }
    if mtda is not None and mtda.keyboard is not None:
        input = request.args.get('input', '', type=str)
        if len(input) > 1:
            if input in map:
                map[input]()
        else:
            mtda.keyboard.write(input)
    return ''


@app.route('/power-toggle')
def power_toggle():
    sid = session_id()
    mtda = app.config['mtda']
    if mtda is not None:
        return mtda.target_toggle(session=sid)
    return ''


def session_id():
    sid = None
    if 'id' in session:
        sid = session['id']
    return sid


class Service:
    def __init__(self, mtda):
        self.mtda = mtda
        self._host = CONSTS.DEFAULTS.WWW_HOST
        self._port = CONSTS.DEFAULTS.WWW_PORT
        app.config['SECRET_KEY'] = secrets.token_hex(16)
        app.config['mtda'] = mtda

    def configure(self, conf):
        if 'host' in conf:
            self._host = conf['host']
        if 'port' in conf:
            self._port = int(conf['port'])

    @property
    def host(self):
        return self._host

    def notify(self, what, event):
        if what == CONSTS.EVENTS.POWER:
            socket.emit("power-event", {"event": event}, namespace="/mtda")
        elif what == CONSTS.EVENTS.SESSION:
            socket.emit("session-event", {"event": event}, namespace="/mtda")
        elif what == CONSTS.EVENTS.STORAGE:
            socket.emit("storage-event", {"event": event}, namespace="/mtda")

    @property
    def port(self):
        return self._port

    def run(self):
        return socket.run(app, debug=False, use_reloader=False,
                          port=self._port, host=self._host)

    def start(self):
        self._thread = threading.Thread(target=self.run)
        return self._thread.start()

    def stop(self):
        socket.stop()

    def write(self, topic, data):
        if topic == CONSTS.CHANNEL.CONSOLE:
            socket.emit("console-output", {"output": data}, namespace="/mtda")



# ---------------------------------------------------------------------------
# Console output for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import collections
import threading


class ConsoleOutput:

    def __init__(self, screen):
        self.exiting = False
        self.rx_alive = False
        self.rx_lock = threading.Lock()
        self.rx_paused = False
        self.rx_queue = collections.deque(maxlen=1000)
        self.rx_thread = None
        self.screen = screen

    def on_event(self, event):
        self.screen.on_event(event)

    def _pause(self):
        self.rx_paused = True

    def pause(self):
        with self.rx_lock:
            self._pause()

    def print(self, data):
        self.screen.print(data)

    def reader(self):
        return None

    def _resume(self):
        while len(self.rx_queue) > 0:
            data = self.rx_queue.popleft()
            self.print(data)
        self.rx_paused = False

    def resume(self):
        with self.rx_lock:
            self._resume()

    def start(self):
        self.rx_alive = True
        self.rx_thread = threading.Thread(
            target=self.reader, name='console_rx')
        self.rx_thread.daemon = True
        self.rx_thread.start()

    def stop(self):
        self.exiting = True

    def toggle(self):
        with self.rx_lock:
            if self.rx_paused is True:
                self._resume()
            else:
                self._pause()

    def write(self, data):
        with self.rx_lock:
            if self.rx_paused is False:
                self.print(data)
            else:
                self.rx_queue.append(data)



# ---------------------------------------------------------------------------
# QEMU storage driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import os
import pathlib

# Local imports
import mtda.constants as CONSTS
from mtda.storage.helpers.image import Image


class QemuController(Image):

    def __init__(self, mtda):
        super().__init__(mtda)
        self.file = "usb-storage.img"
        self.id = None
        self.name = "usb-storage"
        self.mode = CONSTS.STORAGE.ON_TARGET
        self.qemu = mtda.power_controller

    """ Configure this storage controller from the provided configuration"""
    def configure(self, conf):
        self.mtda.debug(3, "storage.qemu.configure()")
        self.lock.acquire()

        result = None
        if 'file' in conf:
            self.file = os.path.realpath(conf['file'])
        d = os.path.dirname(self.file)
        os.makedirs(d, mode=0o755, exist_ok=True)
        if os.path.exists(self.file) is False:
            sparse = pathlib.Path(self.file)
            sparse.touch()
            os.truncate(str(sparse), 8*1024*1024*1024)
        if 'name' in conf:
            self.name = conf['name']

        self.mtda.debug(3, "storage.qemu.configure(): %s" % str(result))
        self.lock.release()
        return result

    def _add(self):
        self.mtda.debug(3, "storage.qemu._add()")

        result = True
        if self.id is None:
            self.id = self.qemu.usb_add(self.name, self.file)
            if self.id is None:
                self.mtda.debug(1, "storage.qemu._add(): "
                                   "usb storage could not be added!")
                result = False

        self.mtda.debug(3, "storage.qemu._add(): %s" % str(result))
        return result

    def _rm(self):
        self.mtda.debug(3, "storage.qemu._rm()")

        result = True
        if self.id is not None:
            result = self.qemu.usb_rm(self.name)
            if result is True:
                self.id = None
            else:
                self.mtda.debug(1, "storage.qemu._rm(): "
                                   "usb storage could not be removed!")

        self.mtda.debug(3, "storage.qemu._rm(): %s" % str(result))
        return result

    """ Get file used by the USB Function driver"""
    def probe(self):
        self.mtda.debug(3, "storage.qemu.probe()")
        self.lock.acquire()

        result = self.qemu.variant == "qemu"
        if result is False:
            self.mtda.debug(1, "storage.qemu.probe(): "
                               "qemu power controller required!")

        self.mtda.debug(3, "storage.qemu.probe(): %s" % str(result))
        self.lock.release()
        return result

    def supports_hotplug(self):
        return True

    """ Attach the shared storage device to the host"""
    def to_host(self):
        self.mtda.debug(3, "storage.qemu.to_host()")
        self.lock.acquire()

        result = self._rm()
        if result is True:
            self.mode = CONSTS.STORAGE.ON_HOST

        self.mtda.debug(3, "storage.qemu.to_host(): %s" % str(result))
        self.lock.release()
        return result

    """ Attach the shared storage device to the target"""
    def to_target(self):
        self.mtda.debug(3, "storage.qemu.to_target()")
        self.lock.acquire()

        result = self._close()
        if result is True:
            result = self._umount()
        if result is True:
            result = self._add()
            if result is True:
                self.mode = CONSTS.STORAGE.ON_TARGET

        self.mtda.debug(3, "storage.qemu.to_target(): %s" % str(result))
        self.lock.release()
        return result

    """ Determine where the shared storage device is attached"""
    def _status(self):
        self.mtda.debug(3, "storage.qemu.status()")

        result = self.mode

        self.mtda.debug(3, "storage.qemu.status(): %s" % str(result))
        return result


def instantiate(mtda):
    return QemuController(mtda)



# ---------------------------------------------------------------------------
# Storage driver for Samsung sdmux on MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import subprocess

# Local imports
import mtda.constants as CONSTS
from mtda.storage.helpers.image import Image


class SamsungSdMuxStorageController(Image):

    def __init__(self, mtda):
        super().__init__(mtda)
        self.file = "/dev/sda"
        self.serial = "sdmux"

    """ Configure this storage controller from the provided configuration"""
    def configure(self, conf):
        self.mtda.debug(3, "storage.samsung.configure()")

        result = None
        if 'device' in conf:
            self.file = conf['device']
        if 'serial' in conf:
            self.serial = conf['serial']

        self.mtda.debug(3, "storage.samsung.configure(): %s" % str(result))
        return result

    """ Check presence of the sdmux"""
    def probe(self):
        self.mtda.debug(3, "storage.samsung.probe()")

        result = True
        try:
            subprocess.check_output([
                "sd-mux-ctrl", "-e", self.serial, "-i"
            ])
        except subprocess.CalledProcessError:
            result = False

        self.mtda.debug(3, "storage.samsung.probe(): %s" % str(result))
        return result

    """ Attach the SD card to the host"""
    def to_host(self):
        self.mtda.debug(3, "storage.samsung.to_host()")

        result = True
        try:
            subprocess.check_output([
                "sd-mux-ctrl", "-e", self.serial, "--ts"
            ])
        except subprocess.CalledProcessError:
            result = False

        self.mtda.debug(3, "storage.samsung.to_host(): %s" % str(result))
        return result

    """ Attach the SD card to the target"""
    def to_target(self):
        self.mtda.debug(3, "storage.samsung.to_target()")
        self.lock.acquire()

        result = self._close()
        if result is True:
            result = self._umount()
        if result is True:
            try:
                subprocess.check_output([
                    "sd-mux-ctrl", "-e", self.serial, "--dut"
                ])
            except subprocess.CalledProcessError:
                result = False

        self.mtda.debug(3, "storage.samsung.to_target(): %s" % str(result))
        self.lock.release()
        return result

    """ Determine where is the SD card attached"""
    def _status(self):
        self.mtda.debug(3, "storage.samsung.status()")

        try:
            status = subprocess.check_output([
                "sd-mux-ctrl", "-e", self.serial, "-u"
            ]).decode("utf-8").splitlines()
            result = CONSTS.STORAGE.UNKNOWN
            for s in status:
                if s == "SD connected to: TS":
                    result = CONSTS.STORAGE.ON_HOST
                    break
                elif s == "SD connected to: DUT":
                    result = CONSTS.STORAGE.ON_TARGET
                    break
        except subprocess.CalledProcessError:
            self.mtda.debug(1, "storage.samsung.status(): sd-mux-ctrl failed!")
            result = CONSTS.STORAGE.UNKNOWN

        self.mtda.debug(3, "storage.samsung.status(): %s" % str(result))
        return result


def instantiate(mtda):
    return SamsungSdMuxStorageController(mtda)



# ---------------------------------------------------------------------------
# Telnet console driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import socket
import time
from telnetlib import Telnet

# Local imports
from mtda.console.interface import ConsoleInterface


class TelnetConsole(ConsoleInterface):

    def __init__(self, mtda):
        self.telnet = None
        self.host = "localhost"
        self.mtda = mtda
        self.port = 23
        self.opened = False
        self.delay = 5
        self.timeout = 10

    """ Configure this console from the provided configuration"""
    def configure(self, conf, role='console'):
        self.mtda.debug(3, "console.telnet.configure()")

        if 'host' in conf:
            self.host = conf['host']
        if 'port' in conf:
            self.port = conf['port']
        if 'delay' in conf:
            self.delay = conf['delay']
        if 'timeout' in conf:
            self.timeout = conf['timeout']

    def probe(self):
        self.mtda.debug(3, "console.telnet.probe()")
        result = True
        self.mtda.debug(3, "console.telnet.probe(): %s" % str(result))
        return result

    def open(self):
        self.mtda.debug(3, "console.telnet.open()")

        result = self.opened
        if self.opened is False:
            try:
                self.telnet = Telnet()
                self.telnet.open(self.host, self.port, self.timeout)
                self.opened = True
                result = True
            except OSError:
                result = False

        self.mtda.debug(3, "console.telnet.open(): %s" % str(result))
        return result

    def close(self):
        self.mtda.debug(3, "console.telnet.close()")

        if self.opened is True:
            self.opened = False
            self.telnet.get_socket().shutdown(socket.SHUT_WR)
            self.telnet.close()
            self.telnet = None
        result = (self.opened is False)

        self.mtda.debug(3, "console.telnet.close(): %s" % str(result))
        return result

    """ Return number of pending bytes to read"""
    def pending(self):
        self.mtda.debug(3, "console.telnet.pending()")

        if self.opened is True:
            result = self.telnet.sock_avail()
        else:
            result = False

        self.mtda.debug(3, "console.telnet.pending(): %s" % str(result))
        return result

    """ Read bytes from the console"""
    def read(self, n=1):
        self.mtda.debug(3, "console.telnet.read(n=%d)" % n)

        if self.opened is False:
            time_before_open = time.time()
            self.open()
            if self.opened is False:
                # make sure we do not return too quickly
                # if we could not connect
                self.mtda.debug(4, "console.telnet.read():"
                                " failed to connnect!")
                time_after_open = time.time()
                elapsed_time = time_after_open - time_before_open
                if elapsed_time < self.delay:
                    delay = self.delay - elapsed_time
                    self.mtda.debug(4, "console.telnet.read():"
                                    " sleeping {0} seconds".format(delay))
                    time.sleep(delay)

        data = bytearray()
        try:
            while n > 0 and self.opened is True:
                avail = self.telnet.read_some()
                data = data + avail
                n = n - len(avail)
        except socket.timeout:
            self.mtda.debug(2, "console.telnet.read(): timeout!")

        self.mtda.debug(3, "console.telnet.read(): %d bytes" % len(data))
        return data

    """ Write to the console"""
    def write(self, data):
        self.mtda.debug(3, "console.telnet.write()")

        if self.opened is True:
            result = self.telnet.write(data)
        else:
            self.mtda.debug(2, "console.telnet.write(): not connected!")
            result = None

        self.mtda.debug(3, "console.telnet.write(): %s" % str(result))
        return result


def instantiate(mtda):
    return TelnetConsole(mtda)



# ---------------------------------------------------------------------------
# MTDA main
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import configparser
import gevent
import glob
import importlib
import os
import queue
import socket
import sys
import threading
import time
import zmq

# Local imports
from mtda.console.input import ConsoleInput
from mtda.console.logger import ConsoleLogger
from mtda.console.remote import RemoteConsole, RemoteMonitor
from mtda.storage.writer import AsyncImageWriter
import mtda.constants as CONSTS
import mtda.discovery
import mtda.keyboard.controller
import mtda.power.controller
import mtda.scripts
import mtda.video.controller
import mtda.utils
from mtda import __version__
from mtda.support.usb import Composite

try:
    www_support = True
    import mtda.www
except ModuleNotFoundError:
    www_support = False


_NOPRINT_TRANS_TABLE = {
    i: '.' for i in range(0, sys.maxunicode + 1) if not chr(i).isprintable()
}

DEFAULT_PREFIX_KEY = 'ctrl-a'
DEFAULT_PASTEBIN_EP = "http://pastebin.com/api/api_post.php"


def _make_printable(s):
    return s.translate(_NOPRINT_TRANS_TABLE)


class MultiTenantDeviceAccess:

    def __init__(self):
        self.config_files = ['mtda.ini']
        self.console = None
        self.socket = None
        self.console_logger = None
        self.monitor = None
        self.monitor_logger = None
        self.console_input = None
        self.console_output = None
        self.monitor_output = None
        self.debug_level = 0
        self.env = {}
        self.fuse = False
        self.keyboard = None
        self.video = None
        self.mtda = self
        self.name = socket.gethostname()
        self.assistant = None
        self.power_controller = None
        self.power_on_script = None
        self.power_off_script = None
        self.power_monitors = []
        self.storage_controller = None
        self._pastebin_api_key = None
        self._pastebin_endpoint = None
        self._storage_mounted = False
        self._storage_opened = False
        self._storage_owner = None
        self._writer = None
        self._writer_data = None
        self.blksz = CONSTS.WRITER.READ_SIZE
        self.usb_switches = []
        self.ctrlport = 5556
        self.conport = 5557
        self.prefix_key = self._prefix_key_code(DEFAULT_PREFIX_KEY)
        self.is_remote = False
        self.is_server = False
        self.remote = None
        self._lock_owner = None
        self._lock_expiry = None
        self._power_expiry = None
        self._power_lock = threading.Lock()
        self._session_lock = threading.Lock()
        self._session_timer = None
        self._sessions = {}
        self._socket_lock = threading.Lock()
        self._time_from_pwr = None
        self._time_from_str = None
        self._time_until_str = None
        self._uptime = 0
        self.version = __version__
        self._www = None

        # Config file in $HOME/.mtda/config
        home = os.getenv('HOME', '')
        if home != '':
            self.config_files.append(os.path.join(home, '.mtda', 'config'))

        # Config file in /etc/mtda/config
        if os.path.exists('/etc'):
            self.config_files.append(os.path.join('/etc', 'mtda', 'config'))

    def agent_version(self):
        return self.version

    def command(self, args, session=None):
        self.mtda.debug(3, "main.command()")

        self._session_check(session)
        result = False
        if self.power_locked(session) is False:
            result = self.power_controller.command(args)

        self.mtda.debug(3, "main.command(): %s" % str(result))
        return result

    def _composite_start(self):
        self.mtda.debug(3, "main._composite_start()")

        result = True
        storage = self.storage_controller
        if storage is not None and storage.variant == 'usbf':
            status, _, _ = self.storage_status()
            enabled = status == CONSTS.STORAGE.ON_TARGET
            self.mtda.debug(3, "main._composite_start(): "
                               "with storage? {}".format(enabled))
            Composite.storage_toggle(enabled)
            result = Composite.install()

        self.mtda.debug(3, "main._composite_start(): {}".format(result))
        return result

    def _composite_stop(self):
        self.mtda.debug(3, "main._composite_stop()")

        result = None
        storage = self.storage_controller
        if storage is not None and storage.variant == 'usbf':
            Composite.remove()

        self.mtda.debug(3, "main._composite_stop(): {}".format(result))
        return result

    def config_set_power_timeout(self, timeout, session=None):
        self.mtda.debug(3, "main.config_set_power_timeout()")

        result = self._power_timeout
        self._power_timeout = timeout
        if timeout == 0:
            self._power_expiry = None
        self._session_check()

        self.mtda.debug(3, "main.config_set_power_timeout(): "
                           "{}".format(result))
        return result

    def config_set_session_timeout(self, timeout, session=None):
        self.mtda.debug(3, "main.config_set_session_timeout()")

        if timeout < CONSTS.SESSION.MIN_TIMEOUT:
            timeout = CONSTS.SESSION.MIN_TIMEOUT

        result = self._session_timeout
        self._session_timeout = timeout

        with self._session_lock:
            now = time.monotonic()
            for s in self._sessions:
                left = self._sessions[s] - now
                if left > timeout:
                    self._sessions[s] = now + timeout
        self._session_check()

        self.mtda.debug(3, "main.config_set_session_timeout(): "
                           "{}".format(result))
        return result

    def console_prefix_key(self):
        self.mtda.debug(3, "main.console_prefix_key()")
        return self.prefix_key

    def _prefix_key_code(self, prefix_key):
        prefix_key = prefix_key.lower()
        key_dict = {'ctrl-a': '\x01', 'ctrl-b': '\x02', 'ctrl-c': '\x03',
                    'ctrl-d': '\x04', 'ctrl-e': '\x05', 'ctrl-f': '\x06',
                    'ctrl-g': '\x07', 'ctrl-h': '\x08', 'ctrl-i': '\x09',
                    'ctrl-j': '\x0A', 'ctrl-k': '\x0B', 'ctrl-l': '\x0C',
                    'ctrl-n': '\x0E', 'ctrl-o': '\x0F', 'ctrl-p': '\x10',
                    'ctrl-q': '\x11', 'ctrl-r': '\x12', 'ctrl-s': '\x13',
                    'ctrl-t': '\x14', 'ctrl-u': '\x15', 'ctrl-v': '\x16',
                    'ctrl-w': '\x17', 'ctrl-x': '\x18', 'ctrl-y': '\x19',
                    'ctrl-z': '\x1A'}

        if prefix_key in key_dict:
            key_ascii = key_dict[prefix_key]
            return key_ascii
        else:
            raise ValueError("the prefix key specified '{0}' is not "
                             "supported".format(prefix_key))

    def console_getkey(self):
        self.mtda.debug(3, "main.console_getkey()")
        result = None
        try:
            result = self.console_input.getkey()
        except AttributeError:
            print("Initialize the console using console_init first")
        self.mtda.debug(3, "main.console_getkey(): %s" % str(result))
        return result

    def console_init(self):
        self.console_input = ConsoleInput()
        self.console_input.start()

    def console_clear(self, session=None):
        self.mtda.debug(3, "main.console_clear()")

        self._session_check(session)
        if self.console_locked(session):
            self.mtda.debug(2, "console_clear(): console is locked")
            return None
        if self.console_logger is not None:
            result = self.console_logger.clear()
        else:
            result = None

        self.mtda.debug(3, "main.console_clear(): %s" % str(result))
        return result

    def console_dump(self, session=None):
        self.mtda.debug(3, "main.console_dump()")

        self._session_check(session)
        if self.console_locked(session):
            self.mtda.debug(2, "console_dump(): console is locked")
            return None

        result = None
        if self.console_logger is not None:
            result = self.console_logger.dump()

        self.mtda.debug(3, "main.console_dump(): %s" % str(result))
        return result

    def console_flush(self, session=None):
        self.mtda.debug(3, "main.console_flush()")

        self._session_check(session)
        if self.console_locked(session):
            self.mtda.debug(2, "console_flush(): console is locked")
            return None

        result = None
        if self.console_logger is not None:
            result = self.console_logger.flush()

        self.mtda.debug(3, "main.console_flush(): %s" % str(result))
        return result

    def console_head(self, session=None):
        self.mtda.debug(3, "main.console_head()")

        self._session_check(session)
        result = None
        if self.console_logger is not None:
            result = self.console_logger.head()

        self.mtda.debug(3, "main.console_head(): %s" % str(result))
        return result

    def console_lines(self, session=None):
        self.mtda.debug(3, "main.console_lines()")

        self._session_check(session)
        result = 0
        if self.console_logger is not None:
            result = self.console_logger.lines()

        self.mtda.debug(3, "main.console_lines(): %s" % str(result))
        return result

    def console_locked(self, session=None):
        self.mtda.debug(3, "main.console_locked()")

        self._session_check(session)
        result = self._check_locked(session)

        self.mtda.debug(3, "main.console_locked(): %s" % str(result))
        return result

    def console_print(self, data, session=None):
        self.mtda.debug(3, "main.console_print()")

        self._session_check(session)
        result = None
        if self.console_logger is not None:
            result = self.console_logger.print(data)

        self.mtda.debug(3, "main.console_print(): %s" % str(result))
        return result

    def console_prompt(self, newPrompt=None, session=None):
        self.mtda.debug(3, "main.console_prompt()")

        self._session_check(session)
        result = None
        if self.console_locked(session) is False and \
           self.console_logger is not None:
            result = self.console_logger.prompt(newPrompt)

        self.mtda.debug(3, "main.console_prompt(): %s" % str(result))
        return result

    def console_remote(self, host, screen):
        self.mtda.debug(3, "main.console_remote()")

        result = None
        if self.is_remote is True:
            # Stop previous remote console
            if self.console_output is not None:
                self.console_output.stop()
            if host is not None:
                # Create and start our remote console
                self.console_output = RemoteConsole(host, self.conport, screen)
                self.console_output.start()
            else:
                self.console_output = None

        self.mtda.debug(3, "main.console_remote(): %s" % str(result))
        return result

    def console_run(self, cmd, session=None):
        self.mtda.debug(3, "main.console_run()")

        self._session_check(session)
        result = None
        if self.console_locked(session) is False and \
           self.console_logger is not None:
            result = self.console_logger.run(cmd)

        self.mtda.debug(3, "main.console_run(): %s" % str(result))
        return result

    def console_send(self, data, raw=False, session=None):
        self.mtda.debug(3, "main.console_send()")

        self._session_check(session)
        result = None
        if self.console_locked(session) is False and \
           self.console_logger is not None:
            result = self.console_logger.write(data, raw)

        self.mtda.debug(3, "main.console_send(): %s" % str(result))
        return result

    def console_tail(self, session=None):
        self.mtda.debug(3, "main.console_tail()")

        self._session_check(session)
        if self.console_locked(session) is False and \
           self.console_logger is not None:
            result = self.console_logger.tail()

        self.mtda.debug(3, "main.console_tail(): %s" % str(result))
        return result

    def console_toggle(self, session=None):
        self.mtda.debug(3, "main.console_toggle()")

        result = None
        self._session_check(session)
        if self.console_output is not None:
            self.console_output.toggle()
        if self.monitor_output is not None:
            self.monitor_output.toggle()

        self.mtda.debug(3, "main.console_toggle(): %s" % str(result))
        return result

    def console_wait(self, what, timeout=None, session=None):
        self.mtda.debug(3, "main.console_wait()")

        self._session_check(session)
        result = None
        if session is not None and timeout is None:
            self.mtda.debug(1, "main.console_wait(): no timeout specified!")
            timeout = CONSTS.RPC.TIMEOUT
        if self.console_locked(session) is False and \
           self.console_logger is not None:
            result = self.console_logger.wait(what, timeout)

        self.mtda.debug(3, "main.console_wait(): %s" % str(result))
        return result

    def debug(self, level, msg):
        if self.debug_level >= level:
            if self.debug_level == 0:
                prefix = "# "
            else:
                prefix = "# debug%d: " % level
            msg = str(msg).replace("\n", "\n%s ... " % prefix)
            lines = msg.splitlines()
            sys.stderr.buffer.write(prefix.encode("utf-8"))
            for line in lines:
                sys.stderr.buffer.write(_make_printable(line).encode("utf-8"))
                sys.stderr.buffer.write(b"\n")
                sys.stderr.buffer.flush()

    def env_get(self, name, default=None, session=None):
        self.mtda.debug(3, "env_get()")

        result = default
        if name in self.env:
            result = self.env[name]

        self.mtda.debug(3, "env_get(): %s" % str(result))
        return result

    def env_set(self, name, value, session=None):
        self.mtda.debug(3, "env_set()")

        result = None

        if name in self.env:
            old_value = self.env[name]
            result = old_value
        else:
            old_value = value

        self.env[name] = value
        self.env["_%s" % name] = old_value

        self.mtda.debug(3, "env_set(): %s" % str(result))
        return result

    def keyboard_write(self, input_str, session=None):
        self.mtda.debug(3, "main.keyboard_write()")

        self._session_check(session)
        result = None
        if self.keyboard is not None:
            result = self.keyboard.write(input_str)

        self.mtda.debug(3, "main.keyboard_write(): %s" % str(result))
        return result

    def monitor_remote(self, host, screen):
        self.mtda.debug(3, "main.monitor_remote()")

        result = None
        if self.is_remote is True:
            # Stop previous remote console
            if self.monitor_output is not None:
                self.monitor_output.stop()
            if host is not None:
                # Create and start our remote console in paused
                # (i.e. buffering) state
                self.monitor_output = RemoteMonitor(host, self.conport, screen)
                self.monitor_output.pause()
                self.monitor_output.start()
            else:
                self.monitor_output = None

        self.mtda.debug(3, "main.monitor_remote(): %s" % str(result))
        return result

    def monitor_send(self, data, raw=False, session=None):
        self.mtda.debug(3, "main.monitor_send()")

        self._session_check(session)
        result = None
        if self.console_locked(session) is False and \
           self.monitor_logger is not None:
            result = self.monitor_logger.write(data, raw)

        self.mtda.debug(3, "main.monitor_send(): %s" % str(result))
        return result

    def monitor_wait(self, what, timeout=None, session=None):
        self.mtda.debug(3, "main.monitor_wait()")

        self._session_check(session)
        result = None
        if session is not None and timeout is None:
            self.mtda.debug(1, "main.monitor_wait(): no timeout specified!")
            timeout = CONSTS.RPC.TIMEOUT
        if self.console_locked(session) is False and \
           self.monitor_logger is not None:
            result = self.monitor_logger.wait(what, timeout)

        self.mtda.debug(3, "main.monitor_wait(): %s" % str(result))
        return result

    def pastebin_api_key(self):
        self.mtda.debug(3, "main.pastebin_api_key()")
        return self._pastebin_api_key

    def pastebin_endpoint(self):
        self.mtda.debug(3, "main.pastebin_endpoint()")
        return self._pastebin_endpoint

    def power_locked(self, session=None):
        self.mtda.debug(3, "main.power_locked()")

        self._session_check(session)
        if self.power_controller is None:
            result = True
        else:
            result = self._check_locked(session)

        self.mtda.debug(3, "main.power_locked(): %s" % str(result))
        return result

    def publish(self, topic, data):
        if self.socket is not None:
            with self._socket_lock:
                self.socket.send(topic, flags=zmq.SNDMORE)
                self.socket.send(data)

    def _storage_event(self, status):
        self.notify(CONSTS.EVENTS.STORAGE, status)

    def storage_bytes_written(self, session=None):
        self.mtda.debug(3, "main.storage_bytes_written()")

        self._session_check(session)
        result = self._writer.written

        self.mtda.debug(3, "main.storage_bytes_written(): %s" % str(result))
        return result

    def storage_compression(self, compression, session=None):
        self.mtda.debug(3, "main.storage_compression()")

        self._session_check(session)
        if self.storage_controller is None:
            result = None
        else:
            result = self._writer.compression.value
            self._writer.compression = compression

        self.mtda.debug(3, "main.storage_compression(): %s" % str(result))
        return result

    def storage_close(self, session=None):
        self.mtda.debug(3, "main.storage_close()")

        self._session_check(session)
        if self.storage_controller is None:
            result = False
        else:
            self._writer.stop()
            self._writer_data = None
            self._storage_opened = not self.storage_controller.close()
            self._storage_owner = None
            result = (self._storage_opened is False)

        self.mtda.debug(3, "main.storage_close(): %s" % str(result))
        return result

    def storage_locked(self, session=None):
        self.mtda.debug(3, "main.storage_locked()")

        self._session_check(session)
        if self._check_locked(session):
            result = True
        # Cannot swap the shared storage device between the host and target
        # without a driver
        elif self.storage_controller is None:
            self.mtda.debug(4, "storage_locked(): no shared storage device")
            result = True
        # If hotplugging is supported, swap only if the shared storage
        # isn't opened
        elif self.storage_controller.supports_hotplug() is True:
            result = self._storage_opened
        # We also need a power controller to be safe
        elif self.power_controller is None:
            self.mtda.debug(4, "storage_locked(): no power controller")
            result = True
        # The target shall be OFF
        elif self.target_status() != "OFF":
            self.mtda.debug(4, "storage_locked(): target isn't off")
            result = True
        # Lastly, the shared storage device shall not be opened
        elif self._storage_opened is True:
            self.mtda.debug(4, "storage_locked(): "
                               "shared storage is in used (opened)")
            result = True
        # We may otherwise swap our shared storage device
        else:
            result = False

        self.mtda.debug(3, "main.storage_locked(): %s" % str(result))
        return result

    def storage_mount(self, part=None, session=None):
        self.mtda.debug(3, "main.storage_mount()")

        self._session_check(session)
        if self._storage_mounted is True:
            self.mtda.debug(4, "storage_mount(): already mounted")
            result = True
        elif self.storage_controller is None:
            self.mtda.debug(4, "storage_mount(): no shared storage device")
            return False
        else:
            result = self.storage_controller.mount(part)
            self._storage_mounted = (result is True)

        self.mtda.debug(3, "main.storage_mount(): %s" % str(result))
        return result

    def storage_update(self, dst, offset, session=None):
        self.mtda.debug(3, "main.storage_update()")

        self._session_check(session)
        result = False
        if self.storage_controller is None:
            self.mtda.debug(4, "storage_update(): no shared storage device")
        else:
            try:
                result = self.storage_controller.update(dst, offset)
                if result is True:
                    self._writer.start()
            except (FileNotFoundError, IOError) as e:
                self.mtda.debug(1, "main.storage_update(): "
                                   "%s" % str(e.args[0]))

        self.mtda.debug(3, "main.storage_update(): %s" % str(result))
        return result

    def storage_open(self, session=None):
        self.mtda.debug(3, "main.storage_open()")

        self._session_check(session)
        owner = self._storage_owner
        status, _, _ = self.storage_status()

        if self.storage_controller is None:
            self.mtda.debug(1, "storage_open(): no shared storage device")
            result = False
        elif status != CONSTS.STORAGE.ON_HOST:
            self.mtda.debug(1, "storage_open(): not attached to host")
            result = False
        elif owner is not None and owner != session:
            self.mtda.debug(1, "storage_open(): in use")
            result = False
        else:
            self.storage_close()
            result = self.storage_controller.open()
            if result is True:
                self._storage_owner = session
                self._writer.start()
            self._storage_opened = result

        self.mtda.debug(3, "main.storage_open(): %s" % str(result))
        return result

    def storage_status(self, session=None):
        self.mtda.debug(3, "main.storage_status()")

        self._session_check(session)
        if self.storage_controller is None:
            self.mtda.debug(4, "storage_status(): no shared storage device")
            result = CONSTS.STORAGE.UNKNOWN, False, 0
        else:
            status = self.storage_controller.status()
            result = status, self._writer.writing, self._writer.written

        self.mtda.debug(3, "main.storage_status(): %s" % str(result))
        return result

    def storage_to_host(self, session=None):
        self.mtda.debug(3, "main.storage_to_host()")

        self._session_check(session)
        if self.storage_locked(session) is False:
            result = self.storage_controller.to_host()
            if result is True:
                self._storage_event(CONSTS.STORAGE.ON_HOST)
        else:
            self.mtda.debug(1, "storage_to_host(): shared storage is locked")
            result = False

        self.mtda.debug(3, "main.storage_to_host(): %s" % str(result))
        return result

    def storage_to_target(self, session=None):
        self.mtda.debug(3, "main.storage_to_target()")

        self._session_check(session)
        if self.storage_locked(session) is False:
            self.storage_close()
            result = self.storage_controller.to_target()
            if result is True:
                self._storage_event(CONSTS.STORAGE.ON_TARGET)
        else:
            self.mtda.debug(1, "storage_to_target(): shared storage is locked")
            result = False

        self.mtda.debug(3, "main.storage_to_target(): %s" % str(result))
        return result

    def storage_swap(self, session=None):
        self.mtda.debug(3, "main.storage_swap()")

        self._session_check(session)
        if self.storage_locked(session) is False:
            result, writing, written = self.storage_status(session)
            if result == CONSTS.STORAGE.ON_HOST:
                if self.storage_controller.to_target() is True:
                    self._storage_event(CONSTS.STORAGE.ON_TARGET)
            elif result == CONSTS.STORAGE.ON_TARGET:
                if self.storage_controller.to_host() is True:
                    self._storage_event(CONSTS.STORAGE.ON_HOST)
        result, writing, written = self.storage_status(session)
        return result

        self.mtda.debug(3, "main.storage_swap(): %s" % str(result))
        return result

    def storage_write(self, data, session=None):
        self.mtda.debug(3, "main.storage_write()")

        self._session_check(session)
        if self.storage_controller is None:
            self.mtda.debug(1, "main.storage_write(): no storage")
            result = -1
        elif self._storage_opened is False:
            self.mtda.debug(1, "main.storage_write(): not opened")
            result = -1
        elif self._writer.failed is True:
            self.mtda.debug(1, "main.storage_write(): write error")
            result = -1
        elif session != self._storage_owner:
            self.mtda.debug(1, "main.storage_write(): in use")
            result = -1
        else:
            try:
                if len(data) == 0:
                    self.mtda.debug(2, "main.storage_write(): "
                                       "using queued data")
                    data = self._writer_data
                self._writer_data = data
                self._writer.put(data, timeout=10)
                result = self.blksz
            except queue.Full:
                self.mtda.debug(2, "main.storage_write(): "
                                   "queue is full")
                result = 0

        if self._writer.failed is True:
            self.mtda.debug(1, "main.storage_write(): "
                               "write or decompression error")
            result = -1

        self.mtda.debug(3, "main.storage_write(): %s" % str(result))
        return result

    def systemd_configure(self):
        console = self.console
        storage = self.storage_controller
        video = self.video
        dir = '/lib/systemd/system/mtda.service.d/'
        self.systemd_reset(dir)
        if console is not None and hasattr(console, 'configure_systemd'):
            console.configure_systemd(dir)
        if storage is not None and hasattr(storage, 'configure_systemd'):
            storage.configure_systemd(dir)
        if video is not None and hasattr(video, 'configure_systemd'):
            video.configure_systemd(dir)

    def systemd_reset(self, dir):
        os.makedirs(dir, exist_ok=True)
        deps = glob.glob(os.path.join(dir, 'auto-dep-*.conf'))
        for d in deps:
            os.unlink(d)

    def toggle_timestamps(self):
        self.mtda.debug(3, "main.toggle_timestamps()")

        if self.console_logger is not None:
            result = self.console_logger.toggle_timestamps()
        else:
            print("no console configured/found!", file=sys.stderr)
            result = None

        self.mtda.debug(3, "main.toggle_timestamps(): %s" % str(result))
        return result

    def target_lock(self, session):
        self.mtda.debug(3, "main.target_lock()")

        self._session_check(session)
        with self._session_lock:
            owner = self.target_owner()
            if owner is None or owner == session:
                self._lock_owner = session
                self._session_event("LOCKED %s" % session)
                result = True
            else:
                result = False

        self.mtda.debug(3, "main.target_lock(): %s" % str(result))
        return result

    def target_locked(self, session):
        self.mtda.debug(3, "main.target_locked()")

        self._session_check(session)
        return self._check_locked(session)

    def target_owner(self):
        self.mtda.debug(3, "main.target_owner()")

        return self._lock_owner

    def _power_event(self, status):
        self._power_expiry = None
        if status == CONSTS.POWER.ON:
            self._uptime = time.monotonic()
        elif status == CONSTS.POWER.OFF:
            self._uptime = 0

        for m in self.power_monitors:
            m.power_changed(status)
        self.notify(CONSTS.EVENTS.POWER, status)

    def _env_for_script(self):
        variant = 'unknown'
        if 'variant' in self.env:
            variant = self.env['variant']
        return {
            "env": self.env,
            "mtda": self,
            "scripts": mtda.scripts,
            "sleep": gevent.sleep,
            "variant": variant
        }

    def _load_device_scripts(self):
        env = self._env_for_script()
        mtda.scripts.load_device_scripts(env['variant'], env)
        for e in env.keys():
            setattr(mtda.scripts, e, env[e])

    def _parse_script(self, script):
        self.mtda.debug(3, "main._parse_script()")

        result = None
        if script is not None:
            result = script.replace("... ", "    ")

        self.mtda.debug(3, "main._parse_script(): %s" % str(result))
        return result

    def exec_power_on_script(self):
        self.mtda.debug(3, "main.exec_power_on_script()")

        result = None
        if self.power_on_script:
            self.mtda.debug(4, "exec_power_on_script(): "
                               "%s" % self.power_on_script)
            env = self._env_for_script()
            result = exec(self.power_on_script, env)

        self.mtda.debug(3, "main.exec_power_on_script(): %s" % str(result))
        return result

    def _target_on(self, session=None):
        self.mtda.debug(3, "main._target_on()")

        result = False
        if self.power_locked(session) is False:
            # Toggle the mass storage functions of the usbf controller
            result = self._composite_start()

            # Turn the DUT on
            if result is True:
                result = self.power_controller.on()

            # Resume logging
            if result is True:
                if self.console_logger is not None:
                    self.console_logger.resume()
                if self.monitor_logger is not None:
                    self.monitor_logger.resume()

                # user-provided power-on script may now be executed
                # (target is up and logging running)
                #
                # power sequence:
                #   <power-on>
                #     <power-on-script>
                #       <runtime>
                #     <power-off-script>
                #   <power-off>
                #
                self.exec_power_on_script()

                self._power_event(CONSTS.POWER.ON)

        self.mtda.debug(3, "main._target_on(): {}".format(result))
        return result

    def target_on(self, session=None):
        self.mtda.debug(3, "main.target_on()")

        result = True
        self._session_check(session)
        with self._power_lock:
            status = self._target_status()
            if status != CONSTS.POWER.ON:
                result = False
                if self.power_locked(session) is False:
                    result = self._target_on(session)

        self.mtda.debug(3, "main.target_on(): {}".format(result))
        return result

    def exec_power_off_script(self):
        self.mtda.debug(3, "main.exec_power_off_script()")

        if self.power_off_script:
            env = self._env_for_script()
            exec(self.power_off_script, env)

    def _target_off(self, session=None):
        self.mtda.debug(3, "main._target_off()")

        # call power-off script before anything else
        #
        # power sequence:
        #   <power-on>
        #     <power-on-script>
        #       <runtime>
        #     <power-off-script>
        #   <power-off>
        #
        self.exec_power_off_script()

        # pause console
        if self.console_logger is not None:
            self.console_logger.reset_timer()
            self.console_logger.pause()

        # and monitor
        if self.monitor_logger is not None:
            self.monitor_logger.reset_timer()
            self.monitor_logger.pause()

        # release keyboard
        if self.keyboard is not None:
            self.keyboard.idle()

        result = True
        if self.power_controller is not None:
            result = self.power_controller.off()
        self._composite_stop()
        self._power_event(CONSTS.POWER.OFF)

        self.mtda.debug(3, "main._target_off(): {}".format(result))
        return result

    def target_off(self, session=None):
        self.mtda.debug(3, "main.target_off()")

        result = True
        self._session_check(session)
        with self._power_lock:
            status = self._target_status()
            if status != CONSTS.POWER.OFF:
                result = False
                if self.power_locked(session) is False:
                    result = self._target_off(session)

        self.mtda.debug(3, "main.target_off(): {}".format(result))
        return result

    def _target_status(self, session=None):
        self.mtda.debug(3, "main._target_status()")

        if self.power_controller is None:
            result = CONSTS.POWER.UNSURE
        else:
            result = self.power_controller.status()

        self.mtda.debug(3, "main._target_status(): {}".format(result))
        return result

    def target_status(self, session=None):
        self.mtda.debug(3, "main.target_status()")

        with self._power_lock:
            result = self._target_status(session)

        self.mtda.debug(3, "main.target_status(): {}".format(result))
        return result

    def target_toggle(self, session=None):
        self.mtda.debug(3, "main.target_toggle()")

        result = CONSTS.POWER.UNSURE
        self._session_check(session)
        with self._power_lock:
            if self.power_locked(session) is False:
                status = self._target_status(session)
                if status == CONSTS.POWER.OFF:
                    if self._target_on() is True:
                        result = CONSTS.POWER.ON
                elif status == CONSTS.POWER.ON:
                    if self._target_off() is True:
                        result = CONSTS.POWER.OFF
            else:
                result = CONSTS.POWER.LOCKED

        self.mtda.debug(3, "main.target_toggle(): {}".format(result))
        return result

    def target_unlock(self, session):
        self.mtda.debug(3, "main.target_unlock()")

        result = False
        self._session_check(session)
        with self._session_lock:
            if self.target_owner() == session:
                self._session_event("UNLOCKED %s" % session)
                self._lock_owner = None
                result = True

        self.mtda.debug(3, "main.target_unlock(): %s" % str(result))
        return result

    def target_uptime(self, session=None):
        self.mtda.debug(3, "main.target_uptime()")

        result = 0
        if self._uptime > 0:
            result = time.monotonic() - self._uptime

        self.mtda.debug(3, "main.target_uptime(): %s" % str(result))
        return result

    def usb_find_by_class(self, className, session=None):
        self.mtda.debug(3, "main.usb_find_by_class()")

        self._session_check(session)
        ports = len(self.usb_switches)
        ndx = 0
        while ndx < ports:
            usb_switch = self.usb_switches[ndx]
            if usb_switch.className == className:
                return usb_switch
            ndx = ndx + 1
        return None

    def usb_has_class(self, className, session=None):
        self.mtda.debug(3, "main.usb_has_class()")

        self._session_check(session)
        usb_switch = self.usb_find_by_class(className, session)
        return usb_switch is not None

    def usb_off(self, ndx, session=None):
        self.mtda.debug(3, "main.usb_off()")

        self._session_check(session)
        try:
            if ndx > 0:
                usb_switch = self.usb_switches[ndx-1]
                usb_switch.off()
        except IndexError:
            print("invalid USB switch #" + str(ndx), file=sys.stderr)

    def usb_off_by_class(self, className, session=None):
        self.mtda.debug(3, "main.usb_off_by_class()")

        self._session_check(session)
        usb_switch = self.usb_find_by_class(className, session)
        if usb_switch is not None:
            return usb_switch.off()
        return False

    def usb_on(self, ndx, session=None):
        self.mtda.debug(3, "main.usb_on()")

        self._session_check(session)
        try:
            if ndx > 0:
                usb_switch = self.usb_switches[ndx-1]
                usb_switch.on()
        except IndexError:
            print("invalid USB switch #" + str(ndx), file=sys.stderr)

    def usb_on_by_class(self, className, session=None):
        self.mtda.debug(3, "main.usb_on_by_class()")

        self._session_check(session)
        usb_switch = self.usb_find_by_class(className, session)
        if usb_switch is not None:
            return usb_switch.on()
        return False

    def usb_ports(self, session=None):
        self.mtda.debug(3, "main.usb_ports()")

        self._session_check(session)
        return len(self.usb_switches)

    def usb_status(self, ndx, session=None):
        self.mtda.debug(3, "main.usb_status()")

        self._session_check(session)
        try:
            if ndx > 0:
                usb_switch = self.usb_switches[ndx-1]
                status = usb_switch.status()
                if status == usb_switch.POWERED_OFF:
                    return "OFF"
                elif status == usb_switch.POWERED_ON:
                    return "ON"
                else:
                    return "???"
        except IndexError:
            print("invalid USB switch #" + str(ndx), file=sys.stderr)
            return "ERR"
        return "???"

    def usb_toggle(self, ndx, session=None):
        self.mtda.debug(3, "main.usb_toggle()")

        self._session_check(session)
        try:
            if ndx > 0:
                usb_switch = self.usb_switches[ndx-1]
                usb_switch.toggle()
        except IndexError:
            print("invalid USB switch #" + str(ndx), file=sys.stderr)

    def video_url(self, host="", opts=None):
        self.mtda.debug(3, "main.video_url(host={0}, "
                           "opts={1})".format(host, opts))

        result = None
        if self.video is not None:
            result = self.video.url(host, opts)

        self.mtda.debug(3, "main.video_url(): %s" % str(result))
        return result

    def load_config(self, remote=None, is_server=False, config_files=None):
        self.mtda.debug(3, "main.load_config()")

        if config_files is None:
            config_files = os.getenv('MTDA_CONFIG', self.config_files)

        self.mtda.debug(2, "main.load_config(): "
                           "config_files={}".format(config_files))

        self.remote = remote
        self.is_remote = remote is not None
        self.is_server = is_server
        parser = configparser.ConfigParser()
        configs_found = parser.read(config_files)
        if configs_found is False:
            return

        if parser.has_section('main'):
            self.load_main_config(parser)
        if parser.has_section('pastebin'):
            self.load_pastebin_config(parser)
        if parser.has_section('remote'):
            self.load_remote_config(parser)
        self.load_timeouts_config(parser)
        if parser.has_section('ui'):
            self.load_ui_config(parser)
        if self.is_remote is False and is_server is True:
            if parser.has_section('assistant'):
                self.load_assistant_config(parser)
            if parser.has_section('environment'):
                self.load_environment(parser)
            if parser.has_section('power'):
                self.load_power_config(parser)
            if parser.has_section('console'):
                self.load_console_config(parser)
            if parser.has_section('keyboard'):
                self.load_keyboard_config(parser)
            if parser.has_section('monitor'):
                self.load_monitor_config(parser)
            if parser.has_section('storage'):
                self.load_storage_config(parser)
            if parser.has_section('usb'):
                self.load_usb_config(parser)
            if parser.has_section('video'):
                self.load_video_config(parser)
            if parser.has_section('scripts'):
                scripts = parser['scripts']
                self.power_on_script = self._parse_script(
                    scripts.get('power on', None))
                self.power_off_script = self._parse_script(
                    scripts.get('power off', None))
            else:
                self.power_on_script = self._parse_script(
                    "scripts.power_on()")
                self.power_off_script = self._parse_script(
                    "scripts.power_off()")
            self._load_device_scripts()

            # web-base UI
            if www_support is True:
                self._www = mtda.www.Service(self)
                if parser.has_section('www'):
                    self.load_www_config(parser)

    def load_main_config(self, parser):
        self.mtda.debug(3, "main.load_main_config()")

        # Name of this agent
        self.name = parser.get('main', 'name', fallback=self.name)

        self.mtda.debug_level = int(
            parser.get('main', 'debug', fallback=self.mtda.debug_level))
        self.mtda.fuse = parser.getboolean(
            'main', 'fuse', fallback=self.mtda.fuse)

    def load_environment(self, parser):
        self.mtda.debug(3, "main.load_environment()")

        for opt in parser.options('environment'):
            value = parser.get('environment', opt)
            self.mtda.debug(4, "main.load_environment(): "
                               "%s => %s" % (opt, value))
            self.env_set(opt, value)

    def load_assistant_config(self, parser):
        self.mtda.debug(3, "main.load_assistant_config()")

        try:
            # Get variant
            variant = parser.get('assistant', 'variant')
            # Try loading its support class
            mod = importlib.import_module("mtda.assistant." + variant)
            factory = getattr(mod, 'instantiate')
            self.assistant = factory(self)
            self.assistant.variant = variant
            # Configure the assistant
            self.assistant.configure(dict(parser.items('assistant')))
        except configparser.NoOptionError:
            print('assistant variant not defined!', file=sys.stderr)
        except ImportError:
            print('assistant "%s" could not be found/loaded!' % (
                variant), file=sys.stderr)

    def load_console_config(self, parser):
        self.mtda.debug(3, "main.load_console_config()")

        try:
            # Get variant
            variant = parser.get('console', 'variant')
            # Try loading its support class
            mod = importlib.import_module("mtda.console." + variant)
            factory = getattr(mod, 'instantiate')
            self.console = factory(self)
            self.console.variant = variant
            # Configure the console
            config = dict(parser.items('console'))
            self.console.configure(config)
            timestamps = parser.getboolean('console', 'timestamps',
                                           fallback=None)
            self._time_from_pwr = timestamps
            if timestamps is None or timestamps is True:
                # check 'time-from' / 'time-until' settings if timestamps is
                # either yes or unspecified
                if 'time-until' in config:
                    self._time_until_str = config['time-until']
                    self._time_from_pwr = True
                if 'time-from' in config:
                    self._time_from_str = config['time-from']
                    self._time_from_pwr = False
        except configparser.NoOptionError:
            print('console variant not defined!', file=sys.stderr)
        except ImportError:
            print('console "%s" could not be found/loaded!' % (
                variant), file=sys.stderr)

    def load_keyboard_config(self, parser):
        self.mtda.debug(3, "main.load_keyboard_config()")

        try:
            # Get variant
            variant = parser.get('keyboard', 'variant')
            # Try loading its support class
            mod = importlib.import_module("mtda.keyboard." + variant)
            factory = getattr(mod, 'instantiate')
            self.keyboard = factory(self)
            # Configure the keyboard controller
            self.keyboard.configure(dict(parser.items('keyboard')))
        except configparser.NoOptionError:
            print('keyboard controller variant not defined!', file=sys.stderr)
        except ImportError:
            print('keyboard controller "%s" could not be found/loaded!' % (
                variant), file=sys.stderr)

    def load_monitor_config(self, parser):
        self.mtda.debug(3, "main.load_monitor_config()")

        try:
            # Get variant
            variant = parser.get('monitor', 'variant')
            # Try loading its support class
            mod = importlib.import_module("mtda.console." + variant)
            factory = getattr(mod, 'instantiate')
            self.monitor = factory(self)
            self.monitor.variant = variant
            # Configure the monitor console
            self.monitor.configure(dict(parser.items('monitor')), 'monitor')
        except configparser.NoOptionError:
            print('monitor variant not defined!', file=sys.stderr)
        except ImportError:
            print('monitor "%s" could not be found/loaded!' % (
                variant), file=sys.stderr)

    def load_pastebin_config(self, parser):
        self.mtda.debug(3, "main.load_pastebin_config()")
        self._pastebin_api_key = parser.get('pastebin', 'api-key',
                                            fallback='')
        self._pastebin_endpoint = parser.get('pastebin', 'endpoint',
                                             fallback=DEFAULT_PASTEBIN_EP)

    def load_power_config(self, parser):
        self.mtda.debug(3, "main.load_power_config()")

        try:
            # Get variant
            variant = parser.get('power', 'variant')
            # Try loading its support class
            mod = importlib.import_module("mtda.power." + variant)
            factory = getattr(mod, 'instantiate')
            self.power_controller = factory(self)
            self.power_controller.variant = variant
            # Configure the power controller
            self.power_controller.configure(dict(parser.items('power')))
        except configparser.NoOptionError:
            print('power controller variant not defined!', file=sys.stderr)
        except ImportError:
            print('power controller "%s" could not be found/loaded!' % (
                variant), file=sys.stderr)

    def load_storage_config(self, parser):
        self.mtda.debug(3, "main.load_storage_config()")

        try:
            # Get variant
            variant = parser.get('storage', 'variant')
            # Try loading its support class
            mod = importlib.import_module("mtda.storage." + variant)
            factory = getattr(mod, 'instantiate')
            self.storage_controller = factory(self)
            self.storage_controller.variant = variant
            self._writer = AsyncImageWriter(self, self.storage_controller)
            # Configure the storage controller
            self.storage_controller.configure(dict(parser.items('storage')))
        except configparser.NoOptionError:
            print('storage controller variant not defined!', file=sys.stderr)
        except ImportError:
            print('power controller "%s" could not be found/loaded!' % (
                variant), file=sys.stderr)

    def load_remote_config(self, parser):
        self.mtda.debug(3, "main.load_remote_config()")

        self.conport = int(
            parser.get('remote', 'console', fallback=self.conport))
        self.ctrlport = int(
            parser.get('remote', 'control', fallback=self.ctrlport))
        if self.is_server is False:
            if self.remote is None:
                # Load remote setting from the configuration
                self.remote = parser.get(
                    'remote', 'host', fallback=self.remote)
                # Allow override from the environment
                self.remote = os.getenv('MTDA_REMOTE', self.remote)

            # Attempt to resolve remote using Zeroconf
            watcher = mtda.discovery.Watcher(CONSTS.MDNS.TYPE)
            ip = watcher.lookup(self.remote)
            if ip is not None:
                self.debug(2, "resolved '{}' ({}) "
                              "using Zeroconf".format(self.remote, ip))
                self.remote = ip
        else:
            self.remote = None
        self.is_remote = self.remote is not None

    def load_timeouts_config(self, parser):
        self.mtda.debug(3, "main.load_timeouts_config()")

        result = None
        s = "timeouts"

        self._lock_timeout = int(parser.get(s, "lock",
                                 fallback=CONSTS.DEFAULTS.LOCK_TIMEOUT))
        self._power_timeout = int(parser.get(s, "power",
                                  fallback=CONSTS.DEFAULTS.POWER_TIMEOUT))
        self._session_timeout = int(parser.get(s, "session",
                                    fallback=CONSTS.DEFAULTS.SESSION_TIMEOUT))

        self._lock_timeout = self._lock_timeout * 60
        self._power_timeout = self._power_timeout * 60
        self._session_timeout = self._session_timeout * 60

        self.mtda.debug(3, "main.load_timeouts_config: %s" % str(result))
        return result

    def load_ui_config(self, parser):
        self.mtda.debug(3, "main.load_ui_config()")
        self.prefix_key = self._prefix_key_code(parser.get(
            'ui', 'prefix', fallback=DEFAULT_PREFIX_KEY))

    def load_usb_config(self, parser):
        self.mtda.debug(3, "main.load_usb_config()")

        try:
            # Get number of ports
            usb_ports = int(parser.get('usb', 'ports'))
            for port in range(0, usb_ports):
                port = port + 1
                section = "usb" + str(port)
                if parser.has_section(section):
                    self.load_usb_port_config(parser, section)
        except configparser.NoOptionError:
            usb_ports = 0

    def load_usb_port_config(self, parser, section):
        self.mtda.debug(3, "main.load_usb_port_config()")

        try:
            # Get attributes
            className = parser.get(section, 'class', fallback="")
            variant = parser.get(section, 'variant')

            # Try loading its support class
            mod = importlib.import_module("mtda.usb." + variant)
            factory = getattr(mod, 'instantiate')
            usb_switch = factory(self)

            # Configure and probe the USB switch
            usb_switch.configure(dict(parser.items(section)))
            usb_switch.probe()

            # Store other attributes
            usb_switch.className = className

            # Add this USB switch
            self.usb_switches.append(usb_switch)
        except configparser.NoOptionError:
            print('usb switch variant not defined!', file=sys.stderr)
        except ImportError:
            print('usb switch "%s" could not be found/loaded!' % (
                variant), file=sys.stderr)

    def load_video_config(self, parser):
        self.mtda.debug(3, "main.load_video_config()")

        try:
            # Get variant
            variant = parser.get('video', 'variant')
            # Try loading its support class
            mod = importlib.import_module("mtda.video." + variant)
            factory = getattr(mod, 'instantiate')
            self.video = factory(self)
            # Configure the video controller
            self.video.configure(dict(parser.items('video')))
        except configparser.NoOptionError:
            print('video controller variant not defined!', file=sys.stderr)
        except ImportError:
            print('video controller "%s" could not be found/loaded!' % (
                variant), file=sys.stderr)

    def load_www_config(self, parser):
        self.mtda.debug(3, "main.load_www_config()")

        if www_support is True:
            self._www.configure(dict(parser.items('www')))

    def notify(self, what, info):
        self.mtda.debug(3, "main.notify({},{})".format(what, info))

        result = None
        if self.socket is not None:
            with self._socket_lock:
                self.socket.send(CONSTS.CHANNEL.EVENTS, flags=zmq.SNDMORE)
                self.socket.send_string("{} {}".format(what, info))
        if self._www is not None:
            self._www.notify(what, info)

        self.mtda.debug(3, "main.notify: {}".format(result))
        return result

    def start(self):
        self.mtda.debug(3, "main.start()")

        if self.is_remote is True:
            return True

        # Probe the specified power controller
        if self.power_controller is not None:
            status = self.power_controller.probe()
            if status is False:
                print('Probe of the Power Controller failed!', file=sys.stderr)
                return False

        # Probe the specified storage controller
        if self.storage_controller is not None:
            status = self.storage_controller.probe()
            if status is False:
                print('Probe of the shared storage device failed!',
                      file=sys.stderr)
                return False

        if self.console is not None:
            # Create a publisher
            if self.is_server is True:
                context = zmq.Context()
                socket = context.socket(zmq.PUB)
                socket.bind("tcp://*:%s" % self.conport)
            else:
                socket = None
            self.socket = socket

            # Create and start console logger
            status = self.console.probe()
            if status is False:
                print('Probe of the %s console failed!' % (
                      self.console.variant), file=sys.stderr)
                return False
            self.console_logger = ConsoleLogger(
                self, self.console, socket,
                self.power_controller, b'CON', self._www)
            if self._time_from_str is not None:
                self.console_logger.time_from = self._time_from_str
            if self._time_until_str is not None:
                self.console_logger.time_until = self._time_until_str
            if self._time_from_pwr is not None and self._time_from_pwr is True:
                self.toggle_timestamps()
            self.console_logger.start()

        if self.monitor is not None:
            # Create and start console logger
            status = self.monitor.probe()
            if status is False:
                print('Probe of the %s monitor console failed!' % (
                      self.monitor.variant), file=sys.stderr)
                return False
            self.monitor_logger = ConsoleLogger(
                self, self.monitor, socket, self.power_controller, b'MON')
            self.monitor_logger.start()

        if self.keyboard is not None:
            status = self.keyboard.probe()
            if status is False:
                print('Probe of the %s keyboard failed!' % (
                      self.keyboard.variant), file=sys.stderr)
                return False

        if self.assistant is not None:
            self.power_monitors.append(self.assistant)
            self.assistant.start()

        if self.video is not None:
            status = self.video.probe()
            if status is False:
                print('Probe of the %s video failed!' % (
                      self.video.variant), file=sys.stderr)
                return False
            self.video.start()

        if self._www is not None:
            self._www.start()

        if self.is_server is True:
            handler = self._session_check
            self._session_timer = mtda.utils.RepeatTimer(10, handler)
            self._session_timer.start()

        # Start from a known state
        if self.power_controller is not None:
            self._target_off()
            self.storage_to_target()
        else:
            # Assume the target is ON if we cannot control power delivery
            # and start logging on available console(s)
            if self.console_logger is not None:
                self.console_logger.resume()
            if self.monitor_logger is not None:
                self.monitor_logger.resume()

        return True

    def stop(self):
        self.mtda.debug(3, "main.stop()")

        if self.is_remote is True:
            return True

        # stop sesssion timer
        self._session_timer.cancel()

        # stop web service
        if self._www is not None:
            self._www.stop()

        # stop video
        if self.video is not None:
            self.video.stop()

        # stop assistant
        if self.assistant is not None:
            self.power_monitors.remove(self.assistant)
            self.assistant.stop()

        # stop monitor console
        if self.monitor is not None:
            self.monitor_logger.stop()

        # stop main console
        if self.console is not None:
            self.console_logger.stop()

        if self.socket is not None:
            self.socket.close()
            self.socket = None

    def _session_check(self, session=None):
        self.mtda.debug(3, "main._session_check(%s)" % str(session))

        events = []
        now = time.monotonic()
        power_off = False
        result = None

        with self._session_lock:
            # Register new session
            if session is not None:
                if session not in self._sessions:
                    events.append("ACTIVE %s" % session)
                self._sessions[session] = now + self._session_timeout

            # Check for inactive sessions
            inactive = []
            for s in self._sessions:
                left = self._sessions[s] - now
                self.mtda.debug(3, "session %s: %d seconds" % (s, left))
                if left <= 0:
                    inactive.append(s)
            for s in inactive:
                events.append("INACTIVE %s" % s)
                self._sessions.pop(s, "")

                # Check if we should arm the auto power-off timer
                # i.e. when the last session is removed and a power timeout
                # was set
                if len(self._sessions) == 0 and self._power_timeout > 0:
                    self._power_expiry = now + self._power_timeout
                    self.mtda.debug(2, "device will be powered down in {} "
                                       "seconds".format(self._power_timeout))

            if len(self._sessions) > 0:
                # There are active sessions: reset power expiry
                self._power_expiry = None
            else:
                # Otherwise check if we should auto-power off the target
                if self._power_expiry is not None and now > self._power_expiry:
                    self._lock_expiry = 0
                    power_off = True

            # Release device if the session owning the lock is idle
            if self._lock_owner is not None:
                if session == self._lock_owner:
                    self._lock_expiry = now + self._lock_timeout
                elif now >= self._lock_expiry:
                    events.append("UNLOCKED %s" % self._lock_owner)
                    self._lock_owner = None

        # Send event sessions generated above
        for e in events:
            self._session_event(e)

        # Check if we should auto power-off the device
        if power_off is True:
            self._target_off()
            self.mtda.debug(2, "device powered down after {} seconds of "
                               "inactivity".format(self._power_timeout))

        self.mtda.debug(3, "main._session_check: %s" % str(result))
        return result

    def _session_event(self, info):
        self.mtda.debug(3, "main._session_event(%s)" % str(info))

        result = None
        if info is not None:
            self.notify(CONSTS.EVENTS.SESSION, info)

        self.mtda.debug(3, "main._session_event: %s" % str(result))
        return result

    def _check_locked(self, session):
        self.mtda.debug(3, "main._check_locked()")

        owner = self.target_owner()
        if owner is None:
            return False
        status = False if session == owner else True
        return status

    @property
    def www(self):
        return self._www



# ---------------------------------------------------------------------------
# MTDA Scripts for QEMU
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------


def qemu_boot_from_usb():
    target = "QEMU USB"
    for i in range(60):
        sleep(0.5)
        mtda.console_send('\x1b')
        mtda.console_print('.')
        output = mtda.console_flush()
        if "Boot Manager" in output:
            break
    if "Boot Manager" in output:
        mtda.debug(1, "Entering Boot Manager")
        mtda.console_send('\x1b[B\x1b[B\r')
        sleep(1)
        output = ""
        tries = 10
        mtda.console_clear()
        while target not in output and tries > 0:
            mtda.console_send('\x1b[B')
            sleep(0.5)
            output = mtda.console_flush()
            tries = tries - 1
        if tries > 0:
            mtda.console_send('\r')


def qemu_power_on():
    scripts.check_boot_from_usb()



# ---------------------------------------------------------------------------
# Keyboard interface for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2021 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import abc


class KeyboardController(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def configure(self, conf):
        """ Configure this keyboard controller from the provided
            configuration"""
        return

    @abc.abstractmethod
    def probe(self):
        """ Check presence of the keyboard controller"""
        return

    @abc.abstractmethod
    def idle(self):
        """ Put the keyboard controller in idle state"""
        return

    @abc.abstractmethod
    def backspace(self, repeat=1):
        return False

    @abc.abstractmethod
    def esc(self, repeat=1):
        return False

    @abc.abstractmethod
    def enter(self, repeat=1):
        return False

    @abc.abstractmethod
    def down(self, repeat=1):
        return False

    @abc.abstractmethod
    def left(self, repeat=1):
        return False

    @abc.abstractmethod
    def capsLock(self, repeat=1):
        return False

    @abc.abstractmethod
    def right(self, repeat=1):
        return False

    @abc.abstractmethod
    def up(self, repeat=1):
        return False

    @abc.abstractmethod
    def write(self, str):
        return

    @abc.abstractmethod
    def f1(self, repeat=1):
        return False

    @abc.abstractmethod
    def f2(self, repeat=1):
        return False

    @abc.abstractmethod
    def f3(self, repeat=1):
        return False

    @abc.abstractmethod
    def f4(self, repeat=1):
        return False

    @abc.abstractmethod
    def f5(self, repeat=1):
        return False

    @abc.abstractmethod
    def f6(self, repeat=1):
        return False

    @abc.abstractmethod
    def f7(self, repeat=1):
        return False

    @abc.abstractmethod
    def f8(self, repeat=1):
        return False

    @abc.abstractmethod
    def f9(self, repeat=1):
        return False

    @abc.abstractmethod
    def f10(self, repeat=1):
        return False

    @abc.abstractmethod
    def f11(self, repeat=1):
        return False

    @abc.abstractmethod
    def f12(self, repeat=1):
        return False

    @abc.abstractmethod
    def tab(self, repeat=1):
        return False



# ---------------------------------------------------------------------------
# MTDA Service discovery using Zeroconf
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2021 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import socket
import zeroconf


class ServiceHandlers(object):

    def __init__(self, listener):
        self.listener = listener

    def add_service(self, zeroconf, type, name):
        info = zeroconf.get_service_info(type, name)
        if self.listener.onServiceEvent:
            self.listener.onServiceEvent("ADD", name, info)

    def remove_service(self, zeroconf, type, name):
        if self.listener.onServiceEvent:
            self.listener.onServiceEvent("REMOVE", name)

    def update_service(self, zeroconf, type, name):
        info = zeroconf.get_service_info(type, name)
        if self.listener.onServiceEvent:
            self.listener.onServiceEvent("UPDATE", name, info)


class Watcher(object):

    def __init__(self, domain, onServiceEvent=None):
        self.browser = None
        self.domain = domain
        self.onServiceEvent = onServiceEvent
        all = zeroconf.InterfaceChoice.All
        self.zeroconf = zeroconf.Zeroconf(interfaces=all)

    def listen(self):
        self.browser = zeroconf.ServiceBrowser(
                self.zeroconf, self.domain, ServiceHandlers(self))
        self.browser.run()

    def lookup(self, name):
        result = None

        name = '{}.{}'.format(name, self.domain)
        info = self.zeroconf.get_service_info(self.domain, name)
        if info is not None:
            result = socket.inet_ntoa(info.addresses[0])

        return result

    def shutdown(self):
        if self.browser:
            self.browser.cancel()



# ---------------------------------------------------------------------------
# Storage interface for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import threading

# Local imports
import mtda.constants as CONSTS
from mtda.storage.controller import StorageController


class DockerController(StorageController):

    def __init__(self, mtda):
        self.mtda = mtda
        self._docker = mtda.power_controller
        self._lock = threading.Lock()
        self._mode = CONSTS.STORAGE.ON_TARGET
        self._handle = None

    def close(self):
        self.mtda.debug(3, "storage.docker.close()")

        result = True
        with self._lock:
            if self._handle is not None:
                self._handle = None
                result = self._docker.import_close()

        self.mtda.debug(3, "storage.docker.close(): {}".format(result))
        return result

    def configure(self, conf):
        return True

    def mount(self, part):
        return False

    def open(self):
        self.mtda.debug(3, "storage.docker.open()")

        with self._lock:
            result = False
            if self._mode == CONSTS.STORAGE.ON_HOST:
                self._handle = self._docker.import_open()
                result = self._handle is not None

        self.mtda.debug(3, "storage.docker.open(): {}".format(result))
        return result

    def probe(self):
        self.mtda.debug(3, "storage.docker.probe()")
        self._lock.acquire()

        result = self._docker.variant == "docker"
        if result is False:
            self.mtda.debug(1, "storage.docker.probe(): "
                            "docker power controller required!")

        self.mtda.debug(3, "storage.docker.probe(): {}".format(result))
        self._lock.release()
        return result

    def supports_hotplug(self):
        return False

    def to_host(self):
        self._mode = CONSTS.STORAGE.ON_HOST
        return True

    def to_target(self):
        self._mode = CONSTS.STORAGE.ON_TARGET
        return True

    def status(self):
        return self._mode

    def update(self, dst, offset):
        return False

    def write(self, data):
        self.mtda.debug(3, "storage.docker.write()")

        result = None
        with self._lock:
            if self._handle is not None:
                result = self._handle.write(data)

        self.mtda.debug(3, "storage.docker.write(): {}".format(result))
        return result


def instantiate(mtda):
    return DockerController(mtda)



# ---------------------------------------------------------------------------
# Remote console support for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# Local imports
from mtda.console.output import ConsoleOutput
import mtda.constants as CONSTS

# System imports
import zmq


class RemoteConsole(ConsoleOutput):

    def __init__(self, host, port, screen):
        ConsoleOutput.__init__(self, screen)
        self.context = None
        self.host = host
        self.port = port
        self.socket = None
        self.topic = CONSTS.CHANNEL.CONSOLE

    def connect(self):
        context = zmq.Context()
        socket = context.socket(zmq.SUB)
        socket.connect("tcp://%s:%s" % (self.host, self.port))
        socket.setsockopt(zmq.SUBSCRIBE, self.topic)
        self.context = context
        self.socket = socket
        return socket

    def dispatch(self, topic, data):
        self.write(data)

    def reader(self):
        socket = self.connect()
        try:
            while self.exiting is False:
                topic, data = socket.recv_multipart()
                self.dispatch(topic, data)
        except zmq.error.ContextTerminated:
            self.socket = None

    def stop(self):
        super().stop()
        if self.context is not None:
            self.context.term()
            self.context = None


class RemoteMonitor(RemoteConsole):

    def __init__(self, host, port, screen):
        super().__init__(host, port, screen)
        self.topic = CONSTS.CHANNEL.MONITOR

    def connect(self):
        socket = super().connect()
        socket.setsockopt(zmq.SUBSCRIBE, CONSTS.CHANNEL.EVENTS)
        return socket

    def dispatch(self, topic, data):
        if topic != CONSTS.CHANNEL.EVENTS:
            self.write(data)
        else:
            self.on_event(data.decode("utf-8"))



# ---------------------------------------------------------------------------
# QEMU keyboard driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import time

# Local imports
from mtda.keyboard.controller import KeyboardController


class QemuController(KeyboardController):

    def __init__(self, mtda):
        self.mtda = mtda
        self.qemu = mtda.power_controller

    def configure(self, conf):
        self.mtda.debug(3, "keyboard.qemu.configure()")

    def probe(self):
        self.mtda.debug(3, "keyboard.qemu.probe()")

        result = (self.qemu is not None and self.qemu.variant == "qemu")
        if result is False:
            self.mtda.debug(1, "keyboard.qemu.probe(): "
                               "a qemu power controller is required")

        self.mtda.debug(3, "keyboard.qemu.probe(): %s" % str(result))
        return result

    def idle(self):
        return True

    def press(self, key, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.press()")

        result = True
        while repeat > 0:
            repeat = repeat - 1
            self.qemu.cmd("sendkey %s" % key)
            time.sleep(0.1)
        return result

    def backspace(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.backspace()")
        return self.press("backspace", repeat)

    def enter(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.enter()")
        return self.press("ret", repeat)

    def esc(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.esc()")
        return self.press("esc", repeat)

    def down(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.down()")
        return self.press("down", repeat)

    def left(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.left()")
        return self.press("left", repeat)

    def capsLock(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.capsLock()")
        return self.press("caps_lock", repeat)

    def right(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.right()")
        return self.press("right", repeat)

    def up(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.up()")
        return self.press("up", repeat)

    def f1(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f1()")
        return self.press("f1", repeat)

    def f2(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f2()")
        return self.press("f2", repeat)

    def f3(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f3()")
        return self.press("f3", repeat)

    def f4(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f4()")
        return self.press("f4", repeat)

    def f5(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f5()")
        return self.press("f5", repeat)

    def f6(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f6()")
        return self.press("f6", repeat)

    def f7(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f7()")
        return self.press("f7", repeat)

    def f8(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f8()")
        return self.press("f8", repeat)

    def f9(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f9()")
        return self.press("f9", repeat)

    def f10(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f10()")
        return self.press("f10", repeat)

    def f11(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f11()")
        return self.press("f11", repeat)

    def f12(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.f12()")
        return self.press("f12", repeat)

    def tab(self, repeat=1):
        self.mtda.debug(3, "keyboard.qemu.tab()")
        return self.press("tab", repeat)

    def write(self, str):
        self.mtda.debug(3, "keyboard.qemu.write()")

        for k in str:
            self.press(k)


def instantiate(mtda):
    return QemuController(mtda)



# ---------------------------------------------------------------------------
# Screen output for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import json
import os
import shutil
import sys
import time
import threading


class ScreenOutput:

    def __init__(self, mtda):
        self.lock = threading.Lock()
        self.mtda = mtda
        self.capture_fd = None
        self.capture_time = None

    def capture_data(self, data):
        with self.lock:
            if self.capture_fd is not None:
                delta = time.monotonic_ns() - self.capture_time
                delta = delta / 1000000000.0
                data = json.dumps(data.decode("latin-1"))
                entry = '[%0.6f, "o", %s]\n' % (delta, data)
                self.capture_fd.write(entry)

    def capture_enabled(self):
        return self.capture_fd is not None

    def capture_start(self):
        with self.lock:
            if self.capture_fd is None:
                self.capture_time = time.monotonic_ns()
                fd = open("screen.cast", "w")
                dim = shutil.get_terminal_size((80, 25))
                env = {}
                env["SHELL"] = os.environ["SHELL"]
                env["TERM"] = os.environ["TERM"]
                header = {}
                header["env"] = env
                header["timestamp"] = self.capture_time
                header["version"] = 2
                header["width"] = dim.columns
                header["height"] = dim.lines
                fd.write(json.dumps(header))
                fd.write("\n")
                self.capture_fd = fd

    def capture_stop(self):
        with self.lock:
            if self.capture_fd is not None:
                self.capture_fd.close()
                self.capture_fd = None

    def on_event(self, event):
        return None

    def print(self, data):
        self.write(data)
        self.capture_data(data)

    def write(self, data):
        sys.stdout.buffer.write(data)
        sys.stdout.buffer.flush()



# ---------------------------------------------------------------------------
# HID keyboard driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import os
import select
import time

# Local imports
from mtda.keyboard.controller import KeyboardController
from mtda.support.usb import Composite


class HidKeyboardController(KeyboardController):

    KEY_MOD_LSHIFT = 0x02

    # Timeout (seconds) for writes
    TIMEOUT = 1

    def __init__(self, mtda):
        self.dev = None
        self.fd = None
        self.mtda = mtda
        Composite.mtda = mtda

    def configure(self, conf):
        self.mtda.debug(3, "keyboard.hid.configure()")

        result = Composite.configure('keyboard', conf)
        if 'device' in conf:
            self.dev = conf['device']
            self.mtda.debug(4, "keyboard.hid.configure(): "
                               "will use {}".format(self.dev))

        self.mtda.debug(3, "keyboard.hid.configure(): {}".format(result))
        return result

    def probe(self):
        self.mtda.debug(3, "keyboard.hid.probe()")

        result = True
        if self.dev is None:
            result = False
            self.mtda.debug(1, "keyboard.hid.probe(): "
                               "{} not configured".format(self.dev))

        self.mtda.debug(3, "keyboard.hid.probe(): {}".format(result))
        return result

    def write_report(self, report):
        self.mtda.debug(3, "keyboard.hid.write_report()")

        outs = [self.fd]
        readable, writable, error = select.select([], outs, [], self.TIMEOUT)
        if len(writable) > 0:
            result = self.fd.write(report.encode())
        else:
            result = 0

        self.mtda.debug(3, "keyboard.hid.write_report(): {}".format(result))
        return result

    def idle(self):
        self.mtda.debug(3, "keyboard.hid.idle()")

        result = True
        if self.fd is not None:
            self.fd.close()
            self.fd = None

        self.mtda.debug(3, "keyboard.hid.idle(): {}".format(result))
        return result

    def press(self, key, mod=0x00, repeat=1):
        self.mtda.debug(3, "keyboard.hid.press()")

        if os.path.exists(self.dev) is False:
            self.mtda.debug(1, "keyboard.hid.press(): "
                               "{} not found".format(self.dev))
            return False

        if self.fd is None:
            self.mtda.debug(4, "keyboard.hid.press(): "
                               "opening {}".format(self.dev))
            self.fd = open(self.dev, mode="r+b", buffering=0)

        NULL_CHAR = chr(0)
        result = True
        while repeat > 0:
            repeat = repeat - 1
            try:
                written = self.write_report(
                    chr(mod) + NULL_CHAR + chr(key) + NULL_CHAR * 5)
                time.sleep(0.1)
                self.write_report(NULL_CHAR * 8)
                if repeat > 0:
                    time.sleep(0.1)
                if written < 8:
                    result = False
                    break
            except IOError:
                self.write_report(NULL_CHAR * 8)
                result = False
                break
        return result

    def backspace(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.backspace()")

        return self.press(0x2a, 0, repeat)

    def capsLock(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.capsLock()")

        return self.press(0x39, 0, repeat)

    def enter(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.enter()")

        return self.press(0x28, 0, repeat)

    def esc(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.esc()")

        return self.press(0x29, 0, repeat)

    def down(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.down()")

        return self.press(0x51, 0, repeat)

    def left(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.left()")

        return self.press(0x50, 0, repeat)

    def right(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.right()")

        return self.press(0x4f, 0, repeat)

    def up(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.up()")

        return self.press(0x52, 0, repeat)

    def f1(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f1()")

        return self.press(0x3a, 0, repeat)

    def f2(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f2()")

        return self.press(0x3b, 0, repeat)

    def f3(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f3()")

        return self.press(0x3c, 0, repeat)

    def f4(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f4()")

        return self.press(0x3d, 0, repeat)

    def f5(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f5()")

        return self.press(0x3e, 0, repeat)

    def f6(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f6()")

        return self.press(0x3f, 0, repeat)

    def f7(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f7()")

        return self.press(0x40, 0, repeat)

    def f8(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f8()")

        return self.press(0x41, 0, repeat)

    def f9(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f9()")

        return self.press(0x42, 0, repeat)

    def f10(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f10())")

        return self.press(0x43, 0, repeat)

    def f11(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f11()")

        return self.press(0x44, 0, repeat)

    def f12(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.f12()")

        return self.press(0x45, 0, repeat)

    def tab(self, repeat=1):
        self.mtda.debug(3, "keyboard.hid.tab()")

        return self.press(0x2b, 0, repeat)

    def write(self, what):
        self.mtda.debug(3, "keyboard.hid.write()")

        ret = '\n'
        lower_keys = {
            'a': 0x04, 'b': 0x05, 'c': 0x06, 'd': 0x07, 'e': 0x08, 'f': 0x09,
            'g': 0x0a, 'h': 0x0b, 'i': 0x0c, 'j': 0x0d, 'k': 0x0e, 'l': 0x0f,
            'm': 0x10, 'n': 0x11, 'o': 0x12, 'p': 0x13, 'q': 0x14, 'r': 0x15,
            's': 0x16, 't': 0x17, 'u': 0x18, 'v': 0x19, 'w': 0x1a, 'x': 0x1b,
            'y': 0x1c, 'z': 0x1d, '1': 0x1e, '2': 0x1f, '3': 0x20, '4': 0x21,
            '5': 0x22, '6': 0x23, '7': 0x24, '8': 0x25, '9': 0x26, '0': 0x27,
            ret: 0x28, ' ': 0x2c, '-': 0x2d, '=': 0x2e
        }
        shift_keys = {
            '!': 0x1e, '@': 0x1f, '#': 0x20, '$': 0x21, '%': 0x22, '^': 0x23,
            '&': 0x24, '*': 0x25, '(': 0x26, ')': 0x27, '_': 0x2d, '+': 0x2e
        }
        special_keys = {
            '<down>': 0x51,
            '<enter>': 0x28,
            '<esc>': 0x29,
            '<f1>': 0x3a,
            '<f2>': 0x3b,
            '<f3>': 0x3c,
            '<f4>': 0x3d,
            '<f5>': 0x3e,
            '<f6>': 0x3f,
            '<f7>': 0x40,
            '<f8>': 0x41,
            '<f9>': 0x42,
            '<f10>': 0x43,
            '<f11>': 0x44,
            '<f12>': 0x45,
            '<left>': 0x50,
            '<right>': 0x4f,
            '<up>': 0x52
        }

        if what in special_keys:
            return self.press(special_keys[what])

        for k in what:
            if k in lower_keys:
                self.press(lower_keys[k])
            elif k in shift_keys:
                self.press(shift_keys[k], mod=self.KEY_MOD_LSHIFT)


def instantiate(mtda):
    return HidKeyboardController(mtda)



# ---------------------------------------------------------------------------
# USB Function storage driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import os
import stat

# Local imports
import mtda.constants as CONSTS
from mtda.storage.helpers.image import Image
from mtda.support.usb import Composite


class UsbFunctionController(Image):

    def __init__(self, mtda):
        super().__init__(mtda)
        self.device = None
        self.file = None
        self.mode = CONSTS.STORAGE.ON_HOST
        Composite.mtda = mtda

    """ Configure this storage controller from the provided configuration"""
    def configure(self, conf):
        self.mtda.debug(3, "storage.usbf.configure()")

        result = False
        if 'device' in conf:
            self.device = conf['device']
        if 'file' in conf:
            self.file = conf['file']

        if self.device is not None and self.file is not None:
            self.mtda.debug(1, "storage.usbf.configure(): "
                               "both 'file' ({}) and 'device' ({}) are set, "
                               "using 'file'".format(self.file, self.device))
            self.device = None

        if self.file is None and self.device is not None:
            self.file = self.device

        if self.file is not None:
            result = Composite.configure('storage', conf)

        self.mtda.debug(3, "storage.usbf.configure(): {}".format(result))
        return result

    def configure_systemd(self, dir):
        if self.file is None or os.path.exists(self.file) is False:
            return
        mode = os.stat(self.file).st_mode
        if stat.S_ISBLK(mode) is False:
            return
        device = os.path.basename(self.file)
        dropin = os.path.join(dir, 'auto-dep-storage.conf')
        with open(dropin, 'w') as f:
            f.write('[Unit]\n')
            f.write('Wants=dev-{}.device\n'.format(device))
            f.write('After=dev-{}.device\n'.format(device))

    """ Get file used by the USB Function driver"""
    def probe(self):
        self.mtda.debug(3, "storage.usbf.probe()")

        result = False
        if self.device is not None:
            if os.path.exists(self.device) is True:
                mode = os.stat(self.device).st_mode
                if stat.S_ISBLK(mode) is False:
                    self.mtda.debug(1, "storage.usbf.probe(): "
                                       "{} is not a block "
                                       "device!".format(self.device))
        if self.file is not None:
            if os.path.exists(self.file) is True:
                result = True
            else:
                self.mtda.debug(1, "storage.usbf.probe(): "
                                   "{} not found!".format(self.file))
        else:
            self.mtda.debug(1, "storage.usbf.probe(): "
                               "file not configured!")

        self.mtda.debug(3, "storage.usbf.probe(): {}".format(result))
        return result

    """ Attach the shared storage device to the host"""
    def to_host(self):
        self.mtda.debug(3, "storage.usbf.to_host()")
        self.lock.acquire()

        self.mode = CONSTS.STORAGE.ON_HOST
        result = True

        self.mtda.debug(3, "storage.usbf.to_host(): {}".format(result))
        self.lock.release()
        return result

    """ Attach the shared storage device to the target"""
    def to_target(self):
        self.mtda.debug(3, "storage.usbf.to_target()")
        self.lock.acquire()

        result = self._close()
        if result is True:
            result = self._umount()

        if result is True:
            self.mode = CONSTS.STORAGE.ON_TARGET

        self.lock.release()
        self.mtda.debug(3, "storage.usbf.to_target(): {}".format(result))
        return result

    """ Determine where the shared storage device is attached"""
    def _status(self):
        self.mtda.debug(3, "storage.usbf.status()")

        result = self.mode

        self.mtda.debug(3, "storage.usbf.status(): {}".format(result))
        return result


def instantiate(mtda):
    return UsbFunctionController(mtda)



# ---------------------------------------------------------------------------
# MTDA Client
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2021 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------


import os
import random
import socket
import time
import zerorpc

from mtda.main import MultiTenantDeviceAccess
import mtda.constants as CONSTS


class Client:

    def __init__(self, host=None, session=None, config_files=None):
        agent = MultiTenantDeviceAccess()
        agent.load_config(host, config_files=config_files)
        if agent.remote is not None:
            uri = "tcp://%s:%d" % (agent.remote, agent.ctrlport)
            self._impl = zerorpc.Client(heartbeat=CONSTS.RPC.HEARTBEAT,
                                        timeout=CONSTS.RPC.TIMEOUT)
            self._impl.connect(uri)
        else:
            self._impl = agent
        self._agent = agent

        if session is None:
            HOST = socket.gethostname()
            USER = os.getenv("USER")
            WORDS = "/usr/share/dict/words"
            if os.path.exists(WORDS):
                WORDS = open(WORDS).read().splitlines()
                name = random.choice(WORDS)
                if name.endswith("'s"):
                    name = name.replace("'s", "")
            elif USER is not None and HOST is not None:
                name = "%s@%s" % (USER, HOST)
            else:
                name = "mtda"
            self._session = os.getenv('MTDA_SESSION', name)
        else:
            self._session = session

    def agent_version(self):
        return self._impl.agent_version()

    def config_set_power_timeout(self, timeout):
        return self._impl.config_set_power_timeout(timeout, self._session)

    def config_set_session_timeout(self, timeout):
        return self._impl.config_set_session_timeout(timeout, self._session)

    def console_prefix_key(self):
        return self._agent.console_prefix_key()

    def command(self, args):
        return self._impl.command(args, self._session)

    def console_clear(self):
        return self._impl.console_clear(self._session)

    def console_dump(self):
        return self._impl.console_dump(self._session)

    def console_flush(self):
        return self._impl.console_flush(self._session)

    def console_getkey(self):
        return self._agent.console_getkey()

    def console_init(self):
        return self._agent.console_init()

    def console_head(self):
        return self._impl.console_head(self._session)

    def console_lines(self):
        return self._impl.console_lines(self._session)

    def console_locked(self):
        return self._impl.console_locked(self._session)

    def console_print(self, data):
        return self._impl.console_print(data, self._session)

    def console_prompt(self, newPrompt=None):
        return self._impl.console_prompt(newPrompt, self._session)

    def console_remote(self, host, screen):
        return self._agent.console_remote(host, screen)

    def console_run(self, cmd):
        return self._impl.console_run(cmd, self._session)

    def console_send(self, data, raw=False):
        return self._impl.console_send(data, raw, self._session)

    def console_toggle(self):
        return self._agent.console_toggle(self._session)

    def console_tail(self):
        return self._impl.console_tail(self._session)

    def console_wait(self, what, timeout=None):
        return self._impl.console_wait(what, timeout, self._session)

    def env_get(self, name):
        return self._impl.env_get(name, self._session)

    def env_set(self, name, value):
        return self._impl.env_set(name, value, self._session)

    def keyboard_write(self, data):
        return self._impl.keyboard_write(data, self._session)

    def monitor_remote(self, host, screen):
        return self._agent.monitor_remote(host, screen)

    def monitor_send(self, data, raw=False):
        return self._impl.monitor_send(data, raw, self._session)

    def monitor_wait(self, what, timeout=None):
        return self._impl.monitor_wait(what, timeout, self._session)

    def pastebin_api_key(self):
        return self._agent.pastebin_api_key()

    def pastebin_endpoint(self):
        return self._agent.pastebin_endpoint()

    def power_locked(self):
        return self._impl.power_locked(self._session)

    def storage_bytes_written(self):
        return self._impl.storage_bytes_written(self._session)

    def storage_close(self):
        return self._impl.storage_close(self._session)

    def storage_locked(self):
        return self._impl.storage_locked(self._session)

    def storage_mount(self, part=None):
        return self._impl.storage_mount(part, self._session)

    def storage_open(self):
        tries = 60
        while tries > 0:
            tries = tries - 1
            status = self._impl.storage_open(self._session)
            if status is True:
                return True
            time.sleep(1)
        return False

    def storage_status(self):
        return self._impl.storage_status(self._session)

    def _storage_write(self, image, imgname, imgsize, callback=None):
        # Copy loop
        bytes_wanted = 0
        data = image.read(self._agent.blksz)
        dataread = len(data)
        totalread = 0
        while totalread < imgsize:
            totalread += dataread

            # Report progress via callback
            if callback is not None:
                callback(imgname, totalread, imgsize)

            # Write block to shared storage device
            bytes_wanted = self._impl.storage_write(data, self._session)

            # Check what to do next
            if bytes_wanted < 0:
                break
            elif bytes_wanted > 0:
                # Read next block
                data = image.read(bytes_wanted)
                dataread = len(data)
            else:
                # Agent may continue without further data
                data = b''
                dataread = 0

        # Close the local image
        image.close()

        # Wait for background writes to complete
        while True:
            status, writing, written = self._impl.storage_status(self._session)
            if writing is False:
                break
            if callback is not None:
                callback(imgname, totalread, imgsize)
            time.sleep(0.5)

        # Storage may be closed now
        status = self.storage_close()

        # Provide final update to specified callback
        if status is True and callback is not None:
            callback(imgname, totalread, imgsize)

        # Make sure an error is reported if a write error was received
        if bytes_wanted < 0:
            status = False

        return status

    def storage_update(self, dest, src=None, callback=None):
        path = dest if src is None else src
        imgname = os.path.basename(path)
        try:
            st = os.stat(path)
            imgsize = st.st_size
            image = open(path, "rb")
        except FileNotFoundError:
            return False

        status = self._impl.storage_update(dest, 0, self._session)
        if status is False:
            image.close()
            return False

        self._impl.storage_compression(CONSTS.IMAGE.RAW.value, self._session)
        return self._storage_write(image, imgname, imgsize, callback)

    def storage_write_image(self, path, callback=None):
        # Get size of the (compressed) image
        imgname = os.path.basename(path)

        # Open the specified image
        try:
            st = os.stat(path)
            imgsize = st.st_size
            if path.endswith(".bz2"):
                compression = CONSTS.IMAGE.BZ2.value
            elif path.endswith(".gz"):
                compression = CONSTS.IMAGE.GZ.value
            elif path.endswith(".zst"):
                compression = CONSTS.IMAGE.ZST.value
            elif path.endswith(".xz"):
                compression = CONSTS.IMAGE.XZ.value
            else:
                compression = CONSTS.IMAGE.RAW.value
            self._impl.storage_compression(compression, self._session)
            image = open(path, "rb")
        except FileNotFoundError:
            return False

        # Open the shared storage device
        status = self.storage_open()
        if status is False:
            image.close()
            return False

        return self._storage_write(image, imgname, imgsize, callback)

    def storage_to_host(self):
        return self._impl.storage_to_host(self._session)

    def storage_to_target(self):
        return self._impl.storage_to_target(self._session)

    def storage_swap(self):
        return self._impl.storage_swap(self._session)

    def start(self):
        return self._agent.start()

    def stop(self):
        if self._agent.remote is not None:
            self._impl.close()
        else:
            self._agent.stop()

    def remote(self):
        return self._agent.remote

    def session(self):
        return self._session

    def target_lock(self, retries=0):
        status = False
        while status is False:
            status = self._impl.target_lock(self._session)
            if retries <= 0 or status is True:
                break
            retries = retries - 1
            time.sleep(60)
        return status

    def target_locked(self):
        return self._impl.target_locked(self._session)

    def target_off(self):
        return self._impl.target_off(self._session)

    def target_on(self):
        return self._impl.target_on(self._session)

    def target_status(self):
        return self._impl.target_status(self._session)

    def target_toggle(self):
        return self._impl.target_toggle(self._session)

    def target_unlock(self):
        return self._impl.target_unlock(self._session)

    def target_uptime(self):
        return self._impl.target_uptime(self._session)

    def toggle_timestamps(self):
        return self._impl.toggle_timestamps()

    def usb_find_by_class(self, className):
        return self._impl.usb_find_by_class(className, self._session)

    def usb_has_class(self, className):
        return self._impl.usb_has_class(className, self._session)

    def usb_off(self, ndx):
        return self._impl.usb_off(ndx, self._session)

    def usb_off_by_class(self, className):
        return self._impl.usb_off_by_class(className, self._session)

    def usb_on(self, ndx):
        return self._impl.usb_on(ndx, self._session)

    def usb_on_by_class(self, className):
        return self._impl.usb_on_by_class(className, self._session)

    def usb_ports(self):
        return self._impl.usb_ports(self._session)

    def usb_status(self, ndx):
        return self._impl.usb_status(ndx, self._session)

    def usb_toggle(self, ndx):
        return self._impl.usb_toggle(ndx, self._session)

    def version(self):
        return self._agent.version

    def video_url(self, host="", opts=None):
        if host == "":
            host = os.getenv("MTDA_REMOTE", "")
        return self._impl.video_url(host, opts)



# ---------------------------------------------------------------------------
# Storage interface for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2021 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import abc
import mtda.constants as CONSTS


class StorageController(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def close(self):
        """ Close the storage device"""
        return False

    @abc.abstractmethod
    def configure(self, conf):
        """ Configure this storage controller from the configuration"""
        return True

    @abc.abstractmethod
    def mount(self, part):
        """ Mount the shared storage on the host"""
        return False

    @abc.abstractmethod
    def open(self):
        """ Open the shared storage device for I/O operations"""
        return False

    @abc.abstractmethod
    def probe(self):
        """ Check presence of the shared storage device"""
        return False

    @abc.abstractmethod
    def supports_hotplug(self):
        """ Whether the shared storage device may be hot-plugged"""
        return False

    @abc.abstractmethod
    def to_host(self):
        """ Attach the shared storage device to the host"""
        return False

    @abc.abstractmethod
    def to_target(self):
        """ Attach the shared storage device to the target"""
        return False

    @abc.abstractmethod
    def status(self):
        """ Determine where the shared storage device is attached"""
        return CONSTS.STORAGE.UNKNOWN

    @abc.abstractmethod
    def update(self, dst, offset):
        """ Update specified file"""
        return False

    @abc.abstractmethod
    def write(self, data):
        """ Write data to the shared storage device"""
        return False



# ---------------------------------------------------------------------------
# Assistant interface for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2021 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import abc


class Assistant(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def configure(self, conf):
        """ Configure this assistant from the provided configuration"""
        return

    @abc.abstractmethod
    def probe(self):
        """ Check presence of the assistant"""
        return

    @abc.abstractmethod
    def start(self):
        """ Start the assistant"""
        return



# ---------------------------------------------------------------------------
# USB switch interface for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2021 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import abc


class UsbSwitch(object):
    __metaclass__ = abc.ABCMeta

    POWERED_OFF = 0
    POWERED_ON = 1
    POWERED_UNSURE = 2

    @abc.abstractmethod
    def configure(self, conf):
        """ Configure this USB switch from the provided configuration"""
        return

    @abc.abstractmethod
    def probe(self):
        """ Check presence of the USB switch"""
        return

    @abc.abstractmethod
    def on(self):
        """ Power on the target USB port"""
        return

    @abc.abstractmethod
    def off(self):
        """ Power off the target USB port"""
        return

    @abc.abstractmethod
    def status(self):
        """ Determine the current power state of the USB port"""
        return self.POWERED_UNSURE

    @abc.abstractmethod
    def toggle(self):
        """ Toggle power state of the USB port"""
        return self.POWERED_OFF



# ---------------------------------------------------------------------------
# aviosys power driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import usb.core

# Local imports
from mtda.power.controller import PowerController


class Aviosys8800PowerController(PowerController):

    DEFAULT_USB_VID = 0x067b
    DEFAULT_USB_PID = 0x2303

    def __init__(self, mtda):
        self.dev = None
        self.mtda = mtda
        self.vid = Aviosys8800PowerController.DEFAULT_USB_VID
        self.pid = Aviosys8800PowerController.DEFAULT_USB_PID

    def configure(self, conf):
        """ Configure this power controller from the provided configuration"""
        if 'vid' in conf:
            self.vid = int(conf['vid'], 16)
        if 'pid' in conf:
            self.pid = int(conf['pid'], 16)

    def probe(self):
        self.dev = usb.core.find(idVendor=self.vid, idProduct=self.pid)
        if self.dev is None:
            raise ValueError("Aviosys 8800 device not found!")

    def command(self, args):
        return False

    def on(self):
        """ Power on the attached device"""
        status = self.dev.ctrl_transfer(0x40, 0x01, 0x0001, 0xa0, [])
        return (status == 0)

    def off(self):
        """ Power off the attached device"""
        status = self.dev.ctrl_transfer(0x40, 0x01, 0x0001, 0x20, [])
        return (status == 0)

    def status(self):
        """ Determine the current power state of the attached device"""
        ret = self.dev.ctrl_transfer(0xc0, 0x01, 0x0081, 0x0000, 0x0001)
        if ret[0] == 0xa0:
            return self.POWER_ON
        return self.POWER_OFF


def instantiate(mtda):
    return Aviosys8800PowerController(mtda)



# ---------------------------------------------------------------------------
# QEMU video capture for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import socket

# Local imports
from mtda.video.controller import VideoController


class QemuVideoController(VideoController):

    def __init__(self, mtda):
        self.mtda = mtda
        self.sink = "xvimagesink"
        self.qemu = mtda.power_controller

    def configure(self, conf):
        self.mtda.debug(3, "video.qemu.configure()")

        result = True
        if 'sink' in conf:
            self.sink = conf['sink']

        self.mtda.debug(3, "video.qemu.configure(): " "%s" % str(result))
        return result

    @property
    def format(self):
        return "VNC"

    def probe(self):
        self.mtda.debug(3, "video.qemu.probe()")

        result = self.qemu.variant == "qemu"
        if result is False:
            self.mtda.debug(1, "video.qemu.probe(): "
                            "qemu power controller required!")

        self.mtda.debug(3, "video.qemu.probe(): %s" % str(result))
        return result

    def start(self):
        return True

    def stop(self):
        return True

    def url(self, host="", opts=None):
        self.mtda.debug(3, "video.qemu.url(host={0}, "
                           "opts={1}".format(host, opts))

        if host is None or host == "":
            host = socket.getfqdn()
            self.mtda.debug(3, "video.qemu.url: using host='%s'" % str(host))
        result = 'gst-pipeline: rfbsrc host={0} ! {1}'.format(host, self.sink)
        if opts is not None:
            if 'sink' in opts:
                if 'name' in opts['sink']:
                    result += ' name="{0}"'.format(opts['sink']['name'])

        self.mtda.debug(3, "video.qemu.url(): %s" % str(result))
        return result


def instantiate(mtda):
    return QemuVideoController(mtda)



# ---------------------------------------------------------------------------
# QEMU mass storage support for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import os
import pathlib
import threading

# Local imports
from mtda.usb.switch import UsbSwitch


class QemuMassStorageSwitch(UsbSwitch):

    def __init__(self, mtda):
        self.mtda = mtda
        self.file = "usb-mass-storage.img"
        self.id = None
        self.lock = threading.Lock()
        self.name = "mass-storage"
        self.qemu = mtda.power_controller

    def configure(self, conf):
        self.mtda.debug(3, "usb.qemu_mass_storage.configure()")

        result = True
        self.lock.acquire()
        if 'file' in conf:
            self.file = os.path.realpath(conf['file'])
            if os.path.exists(self.file) is False:
                sparse = pathlib.Path(self.file)
                sparse.touch()
                os.truncate(str(sparse), 8*1024*1024*1024)
        if 'name' in conf:
            self.name = conf['name']

        self.mtda.debug(3, "usb.qemu_mass_storage.configure(): "
                        "%s" % str(result))
        self.lock.release()
        return result

    def probe(self):
        self.mtda.debug(3, "usb.qemu_mass_storage.probe()")

        result = self.qemu.variant == "qemu"
        if result is False:
            self.mtda.debug(1, "usb.qemu_mass_storage.probe(): "
                            "qemu power controller required!")

        self.mtda.debug(3, "usb.qemu_mass_storage.probe(): %s" % str(result))
        return result

    def on(self):
        self.mtda.debug(3, "usb.qemu_mass_storage.on()")

        result = True
        self.lock.acquire()
        if self.id is None:
            self.id = self.qemu.usb_add(self.name, self.file)
            if self.id is None:
                self.mtda.debug(1, "usb.qemu_mass_storage.add(): "
                                "usb storage could not be added!")
                result = False

        self.mtda.debug(3, "usb.qemu_mass_storage.add(): %s" % str(result))
        self.lock.release()
        return result

    def off(self):
        self.mtda.debug(3, "usb.qemu_mass_storage.off()")

        result = True
        self.lock.acquire()
        if self.id is not None:
            result = self.qemu.usb_rm(self.name)
            if result is True:
                self.id = None
            else:
                self.mtda.debug(1, "usb.qemu_mass_storage.off(): "
                                "usb storage could not be removed!")

        self.mtda.debug(3, "usb.qemu_mass_storage.off(): %s" % str(result))
        self.lock.release()
        return result

    def status(self):
        self.mtda.debug(3, "usb.qemu_mass_storage.status()")

        self.lock.acquire()
        result = self.POWERED_ON
        if self.id is None:
            result = self.POWERED_OFF

        self.mtda.debug(3, "usb.qemu_mass_storage.status(): %s" % str(result))
        self.lock.release()
        return result

    def toggle(self):
        self.mtda.debug(3, "usb.qemu_mass_storage.toggle()")
        s = self.status()
        if s == self.POWERED_ON:
            self.off()
            result = self.POWERED_OFF
        else:
            self.on()
            result = self.POWERED_ON

        self.mtda.debug(3, "usb.qemu_mass_storage.toggle(): %s" % str(result))
        return result


def instantiate(mtda):
    return QemuMassStorageSwitch(mtda)



# ---------------------------------------------------------------------------
# start/stop a docker container
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import atexit
import docker
import subprocess
import threading

# Local imports
from mtda.power.controller import PowerController


class DockerPowerController(PowerController):

    def __init__(self, mtda):
        self.mtda = mtda
        self._client = None
        self._command = "sh"
        self._container = None
        self._image = "alpine"
        self._name = "mtda-docker"
        self._lock = threading.Lock()
        self._proc = None

    def configure(self, conf):
        if 'command' in conf:
            self._command = conf['command']
        if 'image' in conf:
            self._image = conf['image']
        if 'name' in conf:
            self._name = conf['name']

    def probe(self):
        self.mtda.debug(3, "power.docker.probe()")

        result = None
        atexit.register(self._stop)
        self._client = docker.from_env()
        # python3-docker version <= 4.4 pulls all tags by default.
        # set tag explicitly to latest if none specified.
        image = self._image.split(":")
        distro = image[0]
        version = image[1] if len(image) > 1 else "latest"
        self._client.images.pull(distro, tag=version)
        result = self._start()

        self.mtda.debug(3, "power.docker.probe(): {}".format(result))
        return result

    def _start(self, image=None):
        self.mtda.debug(3, "power.docker._start({})".format(image))

        if image is None:
            image = self._image

        client = self._client
        result = None

        try:
            container = client.containers.get(self._name)
            if container.status == "running":
                container.stop()
            container.remove()
        except docker.errors.NotFound:
            pass

        self._container = client.containers.create(image,
                                                   stdin_open=True,
                                                   tty=True,
                                                   command=self._command,
                                                   hostname=self._name,
                                                   name=self._name)

        self.mtda.debug(3, "power.docker._start(): {}".format(result))
        return result

    def _stop(self):
        self.mtda.debug(3, "power.docker._stop()")

        result = True
        if self._container is not None:
            result = self._off()
            if result is True:
                cid = self._container.id
                try:
                    self._client.containers.get(cid)
                    self._container.remove()
                    self._container = None
                except docker.errors.NotFound:
                    self.mtda.debug(2, "power.docker._stop(): "
                                    "container {} has vanished".format(cid))
                    pass
                except docker.errors.APIError as e:
                    self.mtda.debug(1, "power.docker._stop(): "
                                    "{}".format(e))
                    result = False

        self.mtda.debug(3, "power.docker._stop(): {}".format(result))
        return result

    def command(self, args):
        return False

    def _import_close(self):
        self.mtda.debug(3, "power.docker._import_close()")

        if self._proc is None:
            raise RuntimeError("no import in progress!")

        result = False
        self._proc.communicate()
        if self._proc.returncode == 0:
            self._start(self._name)
            result = True
        self._proc = None

        self.mtda.debug(3, "power.docker._import_close(): "
                        "{}".format(result))
        return result

    def import_close(self):
        with self._lock:
            return self._import_close()

    def _import_open(self):
        self.mtda.debug(3, "power.docker._import_open()")

        if self._status() != self.POWER_OFF:
            raise RuntimeError("target isn't OFF!")

        self._stop()
        cmd = ['docker', 'import', '-', self._name]
        self._proc = subprocess.Popen(cmd,
                                      stdin=subprocess.PIPE,
                                      stdout=subprocess.PIPE)
        result = self._proc.stdin

        self.mtda.debug(3, "power.docker._import_open(): "
                        "{}".format(result))
        return result

    def import_open(self):
        with self._lock:
            return self._import_open()

    def _on(self):
        self.mtda.debug(3, "power.docker._on()")

        if self._proc is not None:
            raise RuntimeError("import in progress!")

        result = True
        cid = self._container.id
        try:
            self._container.start()
            self._container = self._client.containers.get(cid)
        except docker.errors.NotFound:
            self.mtda.debug(1, "power.docker._on(): "
                            "container {} has vanished".format(cid))
            result = False
        except docker.errors.APIError as e:
            self.mtda.debug(1, "power.docker._on(): "
                            "{}".format(e))
            result = False

        self.mtda.debug(3, "power.docker._on(): {}".format(result))
        return result

    def on(self):
        with self._lock:
            return self._on()

    def _off(self):
        self.mtda.debug(3, "power.docker._off()")

        result = True
        cid = self._container.id
        try:
            self._container.stop()
            self._container = self._client.containers.get(cid)
        except docker.errors.NotFound:
            self.mtda.debug(1, "power.docker._off(): "
                            "container {} has vanished".format(cid))
            result = False
        except docker.errors.APIError as e:
            self.mtda.debug(1, "power.docker._off(): "
                            "{}".format(e))
            result = False

        self.mtda.debug(3, "power.docker._off(): {}".format(result))
        return result

    def off(self):
        with self._lock:
            return self._off()

    def _socket(self):
        self.mtda.debug(3, "power.docker._socket()")

        result = self._container.attach_socket(params={
            'logs': True,
            'stdin': True,
            'stdout': True,
            'stderr': True,
            'stream': True})

        self.mtda.debug(3, "power.docker._socket(): {}".format(result))
        return result

    def socket(self):
        with self._lock:
            return self._socket()

    def _status(self):
        self.mtda.debug(3, "power.docker._status()")

        result = self.POWER_UNSURE
        status = self._container.status
        if status == "running":
            result = self.POWER_ON
        elif status == "created" or status == "exited":
            result = self.POWER_OFF
        else:
            self.mtda.debug(1, "power.docker._status(): "
                            "unknown status: {}".format(status))

        self.mtda.debug(3, "power.docker._status(): {}".format(result))
        return result

    def status(self):
        with self._lock:
            return self._status()


def instantiate(mtda):
    return DockerPowerController(mtda)



# ---------------------------------------------------------------------------
# Serial console driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import os
import serial

# Local imports
from mtda.console.interface import ConsoleInterface


class SerialConsole(ConsoleInterface):

    def __init__(self, mtda):
        self.ser = None
        self.mtda = mtda
        self.port = "/dev/ttyUSB0"
        self.rate = 115200
        self.opened = False
        self.role = 'console'
        self.hotplug = False

    """ Configure this console from the provided configuration"""
    def configure(self, conf, role='console'):
        self.mtda.debug(3, "console.serial.configure()")

        if 'port' in conf:
            self.port = conf['port']
        if 'rate' in conf:
            self.rate = int(conf['rate'], 10)
        self.role = role

    def configure_systemd(self, dir):
        if self.port is None or os.path.exists(self.port) is False:
            return
        device = os.path.basename(self.port)
        dropin = os.path.join(dir, 'auto-dep-{}.conf'.format(self.role))
        with open(dropin, 'w') as f:
            f.write('[Unit]\n')
            f.write('Wants=dev-{}.device\n'.format(device))
            f.write('After=dev-{}.device\n'.format(device))

    def probe(self):
        self.mtda.debug(3, "console.serial.probe()")

        result = True
        if self.hotplug is False:
            result = os.path.exists(self.port)
        if result is True:
            self.ser = serial.Serial()
            self.ser.port = self.port
            self.ser.baudrate = self.rate
        else:
            self.mtda.debug(1, "console.serial.probe(): "
                            "{} does not exist".format(self.port))

        self.mtda.debug(3, "console.serial.probe(): {}".format(result))
        return result

    def open(self):
        self.mtda.debug(3, "console.serial.open()")

        if self.ser is not None:
            if self.opened is False:
                self.ser.open()
                self.opened = True
            else:
                self.mtda.debug(4, "console.serial.open(): already opened")
        else:
            self.mtda.debug(0, "serial console not setup!")

        result = self.opened

        self.mtda.debug(3, "console.serial.open(): %s" % str(result))
        return result

    def close(self):
        self.mtda.debug(3, "console.serial.close()")

        if self.opened is True:
            if self.ser is not None:
                self.ser.cancel_read()
                self.ser.close()
                self.opened = False
            else:
                self.mtda.debug(0, "lost handle of opened serial console!")
        else:
            self.mtda.debug(4, "console.serial.close(): already closed")

        result = self.opened is False
        self.mtda.debug(3, "console.serial.close(): {}".format(result))
        return result

    """ Return number of pending bytes to read"""
    def pending(self):
        self.mtda.debug(3, "console.serial.pending()")

        result = 0
        if self.ser is not None and self.opened is True:
            result = self.ser.inWaiting()

        self.mtda.debug(3, "console.serial.pending(): %s" % str(result))
        return result

    """ Read bytes from the console"""
    def read(self, n=1):
        self.mtda.debug(3, "console.serial.read()")

        result = None
        if self.ser is not None and self.opened is True:
            result = self.ser.read(n)

        self.mtda.debug(3, "console.serial.read(): %s" % str(result))
        return result

    """ Write to the console"""
    def write(self, data):
        self.mtda.debug(3, "console.serial.write(data=%s)" % str(data))

        result = None
        if self.ser is not None and self.opened is True:
            result = self.ser.write(data)

        self.mtda.debug(3, "console.serial.write(): %s" % str(result))
        return result


def instantiate(mtda):
    return SerialConsole(mtda)



# ---------------------------------------------------------------------------
# MTDA Constants
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2021 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

from enum import Enum


class CHANNEL:
    CONSOLE = b'CON'
    EVENTS = b'EVT'
    MONITOR = b'MON'


class DEFAULTS:
    LOCK_TIMEOUT = 5
    POWER_TIMEOUT = 60
    SESSION_TIMEOUT = 5
    WWW_HOST = '127.0.0.1'
    WWW_PORT = 5000


class EVENTS:
    POWER = "POWER"
    SESSION = "SESSION"
    STORAGE = "STORAGE"


class IMAGE(Enum):
    RAW = 0
    BZ2 = 1
    GZ = 2
    ZST = 3
    XZ = 4


class MDNS:
    TYPE = '_MTDA._tcp.local.'


class POWER:
    OFF = "OFF"
    ON = "ON"
    UNSURE = "???"
    LOCKED = "LOCKED"


class RPC:
    HEARTBEAT = 20
    TIMEOUT = 2*60


class SESSION:
    MIN_TIMEOUT = 10


class STORAGE:
    ON_HOST = "HOST"
    ON_TARGET = "TARGET"
    UNKNOWN = "???"


class WRITER:
    QUEUE_SLOTS = 16
    QUEUE_TIMEOUT = 5
    READ_SIZE = 1*1024*1024
    WRITE_SIZE = 1*1024*1024
